# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...core import OutputType, recursive_tile
from ...core.operand import OperandStage
from ...serialization.serializables import KeyField, Int32Field, BoolField
from ...tensor.array_utils import as_same_device, device, cp, sparse
from ...tensor.core import TensorOrder
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin
from ._k_means_common import _execute_merge_update, _relocate_empty_clusters
from ._k_means_elkan import (
    init_bounds_dense,
    init_bounds_sparse,
    update_chunk_dense,
    update_chunk_sparse,
)
from ._k_means_fast import update_center, update_upper_lower_bounds


class KMeansElkanInitBounds(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_ELKAN_INIT_BOUNDS

    _x = KeyField("x")
    _centers = KeyField("centers")
    _center_half_distances = KeyField("center_half_distances")
    _n_clusters = Int32Field("n_clusters")

    def __init__(
        self,
        x=None,
        centers=None,
        center_half_distances=None,
        n_clusters=None,
        sparse=None,
        gpu=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _x=x,
            _centers=centers,
            _center_half_distances=center_half_distances,
            _n_clusters=n_clusters,
            sparse=sparse,
            gpu=gpu,
            _output_types=output_types,
            **kw
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def x(self):
        return self._x

    @property
    def centers(self):
        return self._centers

    @property
    def center_half_distances(self):
        return self._center_half_distances

    @property
    def n_clusters(self):
        return self._n_clusters

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._x = self._inputs[0]
        self._centers = self._inputs[1]
        self._center_half_distances = self._inputs[2]

    @property
    def output_limit(self):
        return 3

    def __call__(self):
        params = []
        # labels
        params.append(
            {
                "shape": (self._x.shape[0],),
                "dtype": np.dtype(np.int32),
                "order": TensorOrder.C_ORDER,
            }
        )
        # upper bounds
        params.append(
            {
                "shape": (self._x.shape[0],),
                "dtype": self._x.dtype,
                "order": TensorOrder.C_ORDER,
            }
        )
        # lower bounds
        params.append(
            {
                "shape": (self._x.shape[0], self._n_clusters),
                "dtype": self._x.dtype,
                "order": TensorOrder.C_ORDER,
            }
        )
        return self.new_tileables(
            [self._x, self._centers, self._center_half_distances], kws=params
        )

    @classmethod
    def tile(cls, op: "KMeansElkanInitBounds"):
        # unify chunks on axis 0
        if has_unknown_shape(op.centers, op.center_half_distances):
            yield
        x = op.x
        centers = yield from recursive_tile(op.centers.rechunk(op.centers.shape))
        center_half_distances = yield from recursive_tile(
            op.center_half_distances.rechunk(op.center_half_distances.shape)
        )

        out_chunks = [list() for _ in range(op.output_limit)]
        for c in x.chunks:
            chunk_op = op.copy().reset_key()
            chunk_params = []
            # labels chunk
            chunk_params.append(
                {
                    "shape": (c.shape[0],),
                    "index": (c.index[0],),
                    "dtype": np.dtype(np.int32),
                    "order": TensorOrder.C_ORDER,
                }
            )
            # upper bounds
            chunk_params.append(
                {
                    "shape": (c.shape[0],),
                    "index": (c.index[0],),
                    "dtype": c.dtype,
                    "order": TensorOrder.C_ORDER,
                }
            )
            # lower bounds
            chunk_params.append(
                {
                    "shape": (c.shape[0], op.n_clusters),
                    "index": (c.index[0], 0),
                    "dtype": c.dtype,
                    "order": TensorOrder.C_ORDER,
                }
            )
            chunks = chunk_op.new_chunks(
                [c, centers.chunks[0], center_half_distances.chunks[0]],
                kws=chunk_params,
            )
            for i, out_chunk in enumerate(chunks):
                out_chunks[i].append(out_chunk)

        out_nsplits = [
            (x.nsplits[0],),
            (x.nsplits[0],),
            (x.nsplits[0], (op.n_clusters,)),
        ]
        out_params = [out.params for out in op.outputs]
        for i, chunks in enumerate(out_chunks):
            out_params[i]["chunks"] = chunks
        for i, nsplits in enumerate(out_nsplits):
            out_params[i]["nsplits"] = nsplits
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=out_params)

    @classmethod
    def execute(cls, ctx, op):
        (x, centers, center_half_distances), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs],
            device=op.device,
            ret_extra=True,
            copy_if_not_writeable=True,
        )

        with device(device_id):
            if xp is cp:  # pragma: no cover
                raise NotImplementedError("cannot support init_bounds for kmeans elkan")

            n_samples = x.shape[0]
            n_clusters = op.n_clusters

            labels = np.full(n_samples, -1, dtype=np.int32)
            upper_bounds = np.zeros(n_samples, dtype=x.dtype)
            lower_bounds = np.zeros((n_samples, n_clusters), dtype=x.dtype)

            if xp is np:
                init_bounds = init_bounds_dense
            else:
                assert xp is sparse
                init_bounds = init_bounds_sparse

            init_bounds(
                x, centers, center_half_distances, labels, upper_bounds, lower_bounds
            )

            ctx[op.outputs[0].key] = labels
            ctx[op.outputs[1].key] = upper_bounds
            ctx[op.outputs[2].key] = lower_bounds


def init_bounds(X, centers, center_half_distances, n_clusters):
    op = KMeansElkanInitBounds(
        x=X,
        centers=centers,
        center_half_distances=center_half_distances,
        n_clusters=n_clusters,
        sparse=False,
        gpu=X.op.gpu,
    )
    return op()


class KMeansElkanUpdate(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_ELKAN_UPDATE

    _x = KeyField("x")
    _sample_weight = KeyField("sample_weight")
    _centers_old = KeyField("centers_old")
    _center_half_distances = KeyField("center_half_distances")
    _distance_next_center = KeyField("distance_next_center")
    _labels = KeyField("labels")
    _upper_bounds = KeyField("upper_bounds")
    _lower_bounds = KeyField("lower_bounds")
    _update_centers = BoolField("update_centers")
    _n_clusters = Int32Field("n_clusters")

    def __init__(
        self,
        x=None,
        sample_weight=None,
        centers_old=None,
        center_half_distances=None,
        distance_next_center=None,
        labels=None,
        upper_bounds=None,
        lower_bounds=None,
        update_centers=None,
        n_clusters=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _x=x,
            _sample_weight=sample_weight,
            _centers_old=centers_old,
            _center_half_distances=center_half_distances,
            _distance_next_center=distance_next_center,
            _labels=labels,
            _upper_bounds=upper_bounds,
            _lower_bounds=lower_bounds,
            _update_centers=update_centers,
            _n_clusters=n_clusters,
            _output_types=output_types,
            **kw
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def x(self):
        return self._x

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def centers_old(self):
        return self._centers_old

    @property
    def center_half_distances(self):
        return self._center_half_distances

    @property
    def distance_next_center(self):
        return self._distance_next_center

    @property
    def labels(self):
        return self._labels

    @property
    def upper_bounds(self):
        return self._upper_bounds

    @property
    def lower_bounds(self):
        return self._lower_bounds

    @property
    def update_centers(self):
        return self._update_centers

    @property
    def n_clusters(self):
        return self._n_clusters

    @property
    def output_limit(self):
        return 5 if self.stage != OperandStage.reduce else 2

    @property
    def _input_fields(self):
        return (
            "_x",
            "_sample_weight",
            "_centers_old",
            "_center_half_distances",
            "_distance_next_center",
            "_labels",
            "_upper_bounds",
            "_lower_bounds",
        )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.stage != OperandStage.reduce:
            input_fields = self._input_fields
            assert len(input_fields) == len(self._inputs)
            inputs_iter = iter(inputs)
            for field in input_fields:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = list(
            (
                # labels
                self._labels.params,
                # upper_bounds
                self._upper_bounds.params,
                # lower_bounds
                self._lower_bounds.params,
            )
        )
        # centers_new
        kws.append(
            {
                "shape": (self._n_clusters, self._x.shape[1]),
                "dtype": self._centers_old.dtype,
                "order": TensorOrder.C_ORDER,
            }
        )
        # weight_in_clusters
        kws.append(
            {
                "shape": (self._n_clusters,),
                "dtype": self._centers_old.dtype,
                "order": TensorOrder.C_ORDER,
            }
        )
        return self.new_tileables(
            [getattr(self, f) for f in self._input_fields], kws=kws
        )

    @classmethod
    def tile(cls, op: "KMeansElkanUpdate"):
        if has_unknown_shape(*op.inputs):
            yield
        x = op.x
        if x.chunk_shape[1] != 1:  # pragma: no cover
            x = yield from recursive_tile(x.rechunk({1: x.shape[1]}))
        sample_weight = yield from recursive_tile(
            op.sample_weight.rechunk({0: x.nsplits[0]})
        )
        labels = yield from recursive_tile(op.labels.rechunk({0: x.nsplits[0]}))
        upper_bounds = yield from recursive_tile(
            op.upper_bounds.rechunk({0: x.nsplits[0]})
        )
        lower_bounds = yield from recursive_tile(
            op.lower_bounds.rechunk({0: x.nsplits[0], 1: op.lower_bounds.shape[1]})
        )
        centers_old = yield from recursive_tile(
            op.centers_old.rechunk(op.centers_old.shape)
        )
        center_half_distances = yield from recursive_tile(
            op.center_half_distances.rechunk(op.center_half_distances.shape)
        )
        distance_next_center = yield from recursive_tile(
            op.distance_next_center.rechunk(op.distance_next_center.shape)
        )

        out_chunks = [list() for _ in range(op.output_limit)]
        for i in range(x.chunk_shape[0]):
            x_chunk = x.cix[i, 0]
            sample_weight_chunk = sample_weight.cix[
                i,
            ]
            labels_chunk = labels.cix[
                i,
            ]
            upper_bounds_chunk = upper_bounds.cix[
                i,
            ]
            lower_bounds_chunk = lower_bounds.cix[i, 0]
            chunk_op = op.copy().reset_key()
            chunk_op.stage = OperandStage.map
            chunk_kws = list(
                (
                    # labels
                    labels_chunk.params,
                    # upper_bounds
                    upper_bounds_chunk.params,
                    # lower_boudns
                    lower_bounds_chunk.params,
                )
            )
            # centers_new
            chunk_kws.append(
                {
                    "index": (0, 0),
                    "shape": (op.n_clusters, x_chunk.shape[1]),
                    "dtype": centers_old.dtype,
                    "order": TensorOrder.C_ORDER,
                }
            )
            # weight_in_clusters
            chunk_kws.append(
                {
                    "index": (0,),
                    "shape": (op.n_clusters,),
                    "dtype": centers_old.dtype,
                    "order": TensorOrder.C_ORDER,
                }
            )
            chunks = chunk_op.new_chunks(
                [
                    x_chunk,
                    sample_weight_chunk,
                    centers_old.chunks[0],
                    center_half_distances.chunks[0],
                    distance_next_center.chunks[0],
                    labels_chunk,
                    upper_bounds_chunk,
                    lower_bounds_chunk,
                ],
                kws=chunk_kws,
            )
            assert len(chunks) == len(out_chunks)
            for oc, c in zip(out_chunks, chunks):
                oc.append(c)

        label_chunks, upper_bounds_chunks, lower_bounds_chunks = out_chunks[:3]
        centers_new_chunks, weight_in_cluster_chunks = out_chunks[3:]

        if op.update_centers:
            # merge centers_new and weight_in_clusters
            merge_op = KMeansElkanUpdate(
                stage=OperandStage.reduce, n_clusters=op.n_clusters
            )
            merge_chunk_kw = [
                centers_new_chunks[0].params,
                weight_in_cluster_chunks[0].params,
            ]
            centers_new_chunk, weight_in_cluster_chunk = merge_op.new_chunks(
                centers_new_chunks + weight_in_cluster_chunks, kws=merge_chunk_kw
            )
        else:
            # the data is meaningless, just pick one
            centers_new_chunk = centers_new_chunks[0]
            weight_in_cluster_chunk = weight_in_cluster_chunks[0]

        out_params = [out.params for out in op.outputs]
        # labels
        out_params[0]["nsplits"] = labels.nsplits
        out_params[0]["chunks"] = label_chunks
        # upper_bounds
        out_params[1]["nsplits"] = upper_bounds.nsplits
        out_params[1]["chunks"] = upper_bounds_chunks
        # lower_bounds
        out_params[2]["nsplits"] = lower_bounds.nsplits
        out_params[2]["chunks"] = lower_bounds_chunks
        # centers_new
        out_params[3]["nsplits"] = tuple((s,) for s in op.outputs[3].shape)
        out_params[3]["chunks"] = [centers_new_chunk]
        # weight_in_clusters
        out_params[4]["nsplits"] = tuple((s,) for s in op.outputs[4].shape)
        out_params[4]["chunks"] = [weight_in_cluster_chunk]
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=out_params)

    @classmethod
    def _execute_reduce(cls, ctx, op):
        return _execute_merge_update(ctx, op)

    @classmethod
    def execute(cls, ctx, op: "KMeansElkanUpdate"):
        if op.stage == OperandStage.reduce:
            return cls._execute_reduce(ctx, op)
        else:
            (
                (
                    x,
                    sample_weight,
                    centers_old,
                    center_half_distances,
                    distance_next_center,
                    labels,
                    upper_bounds,
                    lower_bounds,
                ),
                device_id,
                xp,
            ) = as_same_device(
                [ctx[inp.key] for inp in op.inputs],
                device=op.device,
                ret_extra=True,
                copy_if_not_writeable=True,
            )

            with device(device_id):
                if not op.update_centers:
                    centers_new = centers_old.copy()
                else:
                    centers_new = np.zeros_like(centers_old)
                weight_in_clusters = np.zeros(op.n_clusters, dtype=x.dtype)

                if xp is np:
                    method = update_chunk_dense
                elif xp is sparse:
                    method = update_chunk_sparse
                else:  # pragma: no cover
                    raise NotImplementedError("Does not support run on GPU")

                out_labels, out_upper_bounds, out_lower_bounds = (
                    labels.copy(),
                    upper_bounds.copy(),
                    lower_bounds.copy(),
                )
                method(
                    x,
                    sample_weight,
                    centers_old,
                    center_half_distances,
                    distance_next_center,
                    out_labels,
                    out_upper_bounds,
                    out_lower_bounds,
                    centers_new,
                    weight_in_clusters,
                    op.update_centers,
                )

                # labels
                ctx[op.outputs[0].key] = out_labels
                # upper_bounds
                ctx[op.outputs[1].key] = out_upper_bounds
                # lower_bounds
                ctx[op.outputs[2].key] = out_lower_bounds
                # centers_new
                ctx[op.outputs[3].key] = centers_new
                # weight_in_cluster
                ctx[op.outputs[4].key] = weight_in_clusters


class KMeansElkanPostprocess(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_ELKAN_POSTPROCESS

    _centers_old = KeyField("centers_old")
    _centers_new = KeyField("centers_new")
    _center_shift = KeyField("center_shift")
    _lower_bounds = KeyField("lower_bounds")
    _upper_bounds = KeyField("upper_bounds")
    _labels = KeyField("labels")
    _weight_in_clusters = KeyField("weight_in_clusters")

    def __init__(
        self,
        centers_old=None,
        centers_new=None,
        center_shift=None,
        lower_bounds=None,
        upper_bounds=None,
        labels=None,
        weight_in_clusters=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _centers_old=centers_old,
            _centers_new=centers_new,
            _center_shift=center_shift,
            _lower_bounds=lower_bounds,
            _upper_bounds=upper_bounds,
            _labels=labels,
            _weight_in_clusters=weight_in_clusters,
            _output_types=output_types,
            **kw
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def centers_old(self):
        return self._centers_old

    @property
    def centers_new(self):
        return self._centers_new

    @property
    def center_shift(self):
        return self._center_shift

    @property
    def lower_bounds(self):
        return self._lower_bounds

    @property
    def upper_bounds(self):
        return self._upper_bounds

    @property
    def labels(self):
        return self._labels

    @property
    def weight_in_clusters(self):
        return self._weight_in_clusters

    @property
    def output_limit(self):
        if self.stage is None:
            # for tileable
            return 4
        elif self.stage == OperandStage.combine:
            return 2
        else:
            assert self.stage == OperandStage.reduce
            return 2

    @property
    def _input_fields(self):
        return (
            "_centers_old",
            "_centers_new",
            "_center_shift",
            "_lower_bounds",
            "_upper_bounds",
            "_labels",
            "_weight_in_clusters",
        )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in self._input_fields:
            ob = getattr(self, field)
            if ob is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = [
            # centers_new
            self._centers_new.params,
            # center_shift
            self._center_shift.params,
            # upper_bounds
            self._upper_bounds.params,
            # lower_bounds
            self._lower_bounds.params,
        ]
        return self.new_tileables(
            [getattr(self, f) for f in self._input_fields], kws=kws
        )

    @classmethod
    def tile(cls, op: "KMeansElkanPostprocess"):
        assert len(op.centers_old.chunks) == 1
        assert len(op.centers_new.chunks) == 1
        assert len(op.center_shift.chunks) == 1
        assert len(op._weight_in_clusters.chunks) == 1
        assert op.lower_bounds.chunk_shape[1] == 1

        centers_old_chunk = op.centers_old.chunks[0]
        centers_new_chunk = op.centers_new.chunks[0]
        center_shift_chunk = op.center_shift.chunks[0]
        weight_in_clusters_chunk = op.weight_in_clusters.chunks[0]

        # calculate center shift first
        centers_new_chunk, center_shift_chunk = KMeansElkanPostprocess(
            centers_old=centers_old_chunk,
            centers_new=centers_new_chunk,
            center_shift=center_shift_chunk,
            weight_in_clusters=weight_in_clusters_chunk,
            stage=OperandStage.combine,
        ).new_chunks(
            [
                centers_old_chunk,
                centers_new_chunk,
                center_shift_chunk,
                weight_in_clusters_chunk,
            ],
            kws=[centers_new_chunk.params, center_shift_chunk.params],
        )

        upper_bounds_chunks, lower_bounds_chunks = [], []
        for upper_bound_chunk, lower_bound_chunk, labels_chunk in zip(
            op.upper_bounds.chunks, op.lower_bounds.chunks, op.labels.chunks
        ):
            chunk_kws = [upper_bound_chunk.params, lower_bound_chunk.params]
            upper_bound_chk, lower_bound_chk = KMeansElkanPostprocess(
                center_shift=center_shift_chunk,
                lower_bounds=lower_bound_chunk,
                upper_bounds=upper_bound_chunk,
                labels=labels_chunk,
                stage=OperandStage.reduce,
            ).new_chunks(
                [
                    center_shift_chunk,
                    lower_bound_chunk,
                    upper_bound_chunk,
                    labels_chunk,
                ],
                kws=chunk_kws,
            )
            upper_bounds_chunks.append(upper_bound_chk)
            lower_bounds_chunks.append(lower_bound_chk)

        centers_new_kw = op.centers_new.params
        centers_new_kw["chunks"] = [centers_new_chunk]
        centers_new_kw["nsplits"] = op.centers_new.nsplits
        center_shift_kw = op.center_shift.params
        center_shift_kw["chunks"] = [center_shift_chunk]
        center_shift_kw["nsplits"] = op.center_shift.nsplits
        upper_bounds_kw = op.upper_bounds.params
        upper_bounds_kw["chunks"] = upper_bounds_chunks
        upper_bounds_kw["nsplits"] = op.upper_bounds.nsplits
        lower_bounds_kw = op.lower_bounds.params
        lower_bounds_kw["chunks"] = lower_bounds_chunks
        lower_bounds_kw["nsplits"] = op.lower_bounds.nsplits
        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs,
            kws=[centers_new_kw, center_shift_kw, upper_bounds_kw, lower_bounds_kw],
        )

    @classmethod
    def _execute_combine(cls, ctx, op):
        (
            (centers_old, centers_new, center_shift, weight_in_clusters),
            device_id,
            xp,
        ) = as_same_device(
            [ctx[inp.key] for inp in op.inputs],
            op.device,
            ret_extra=True,
            copy_if_not_writeable=True,
        )

        with device(device_id):
            out_centers_new = centers_new.copy()
            out_center_shift = center_shift.copy()
            update_center(
                centers_old, out_centers_new, out_center_shift, weight_in_clusters
            )

            ctx[op.outputs[0].key] = out_centers_new
            ctx[op.outputs[1].key] = out_center_shift

    @classmethod
    def _execute_reduce(cls, ctx, op):
        (
            (center_shift, lower_bounds, upper_bounds, labels),
            device_id,
            xp,
        ) = as_same_device(
            [ctx[inp.key] for inp in op.inputs],
            op.device,
            ret_extra=True,
            copy_if_not_writeable=True,
        )

        with device(device_id):
            out_upper_bounds = upper_bounds.copy()
            out_lower_bounds = lower_bounds.copy()
            update_upper_lower_bounds(
                out_upper_bounds, out_lower_bounds, labels, center_shift
            )
            ctx[op.outputs[0].key] = out_upper_bounds
            ctx[op.outputs[1].key] = out_lower_bounds

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.combine:
            return cls._execute_combine(ctx, op)
        else:
            assert op.stage == OperandStage.reduce
            return cls._execute_reduce(ctx, op)


def elkan_iter(
    X,
    sample_weight,
    centers_old,
    center_half_distances,
    distance_next_center,
    upper_bounds,
    lower_bounds,
    labels,
    center_shift,
    update_centers=True,
    session=None,
    run_kwargs=None,
):
    update_op = KMeansElkanUpdate(
        x=X,
        sample_weight=sample_weight,
        centers_old=centers_old,
        center_half_distances=center_half_distances,
        distance_next_center=distance_next_center,
        labels=labels,
        upper_bounds=upper_bounds,
        lower_bounds=lower_bounds,
        update_centers=update_centers,
        n_clusters=centers_old.shape[0],
    )
    to_run = []
    ret = update_op()
    to_run.extend(ret)
    labels, upper_bounds, lower_bounds, centers_new, weight_in_clusters = ret

    if update_centers:
        centers_new, weight_in_clusters = _relocate_empty_clusters(
            X,
            sample_weight,
            centers_old,
            centers_new,
            weight_in_clusters,
            labels,
            to_run=to_run,
            session=session,
            run_kwargs=run_kwargs,
        )
        postprocess = KMeansElkanPostprocess(
            centers_old=centers_old,
            centers_new=centers_new,
            center_shift=center_shift,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            labels=labels,
            weight_in_clusters=weight_in_clusters,
        )
        centers_new, center_shift, upper_bounds, lower_bounds = postprocess()

    return (
        centers_new,
        weight_in_clusters,
        upper_bounds,
        lower_bounds,
        labels,
        center_shift,
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import AbstractService
from .core import MetaStoreManagerActor, MetaStoreActor


class MetaSupervisorService(AbstractService):
    """
    Meta service on supervisor.

    Service Configuration
    ---------------------
    {
        "meta" : {
            "store": "<meta store name>",
            # other config related to each store
        }
    }
    """

    async def start(self):
        service_config = self._config["meta"]
        meta_store_name = service_config.get("meta", "dict")
        extra_config = service_config.copy()
        extra_config.pop("meta", None)
        await mo.create_actor(
            MetaStoreManagerActor,
            meta_store_name,
            extra_config,
            uid=MetaStoreManagerActor.default_uid(),
            address=self._address,
        )

    async def stop(self):
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=MetaStoreManagerActor.default_uid(), address=self._address
            )
        )

    async def create_session(self, session_id: str):
        # get MetaStoreManagerActor ref.
        meta_store_manager_ref = await mo.actor_ref(
            self._address, MetaStoreManagerActor.default_uid()
        )
        await meta_store_manager_ref.new_session_meta_store(session_id)

    async def destroy_session(self, session_id: str):
        meta_store_ref = await mo.actor_ref(
            self._address, MetaStoreActor.gen_uid(session_id)
        )
        await mo.destroy_actor(meta_store_ref)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from functools import partial

import numpy as np

from ... import opcodes as OperandDef
from ...core import recursive_tile
from ...core.operand import OperandStage
from ...serialization.serializables import (
    FieldTypes,
    Int32Field,
    ListField,
    StringField,
    BoolField,
    AnyField,
)
from ...utils import flatten, stack_back
from ..core import TensorOrder
from ..operands import (
    TensorOperand,
    TensorMapReduceOperand,
    TensorShuffleProxy,
    TensorOperandMixin,
)
from ..array_utils import as_same_device, device, cp


class PSRSOperandMixin:
    @classmethod
    def preprocess(cls, op, in_data=None):
        if in_data is None:
            in_data = op.inputs[0]
        axis_shape = in_data.shape[op.axis]
        axis_chunk_shape = in_data.chunk_shape[op.axis]

        # rechunk to ensure all chunks on axis have rough same size
        has_unknown_shape = False
        for ns in in_data.nsplits:
            if any(np.isnan(s) for s in ns):
                has_unknown_shape = True
                break

        if not has_unknown_shape:
            axis_chunk_shape = min(axis_chunk_shape, int(np.sqrt(axis_shape)))
            if np.isnan(axis_shape) or any(
                np.isnan(s) for s in in_data.nsplits[op.axis]
            ):
                yield
            chunk_size = int(axis_shape / axis_chunk_shape)
            chunk_sizes = [chunk_size for _ in range(int(axis_shape // chunk_size))]
            if axis_shape % chunk_size > 0:
                chunk_sizes[-1] += axis_shape % chunk_size
            in_data = yield from recursive_tile(
                in_data.rechunk({op.axis: tuple(chunk_sizes)})
            )
            axis_chunk_shape = in_data.chunk_shape[op.axis]

        left_chunk_shape = (
            in_data.chunk_shape[: op.axis] + in_data.chunk_shape[op.axis + 1 :]
        )
        if len(left_chunk_shape) > 0:
            out_idxes = itertools.product(*(range(s) for s in left_chunk_shape))
        else:
            out_idxes = [()]
        # if the size except axis has more than 1, the sorted values on each one may be different
        # another shuffle would be required to make sure each axis except to sort
        # has elements with identical size
        extra_shape = [s for i, s in enumerate(in_data.shape) if i != op.axis]
        if getattr(op, "need_align", None) is None:
            need_align = bool(np.prod(extra_shape, dtype=int) != 1)
        else:
            need_align = op.need_align

        return in_data, axis_chunk_shape, out_idxes, need_align

    @classmethod
    def local_sort_and_regular_sample(
        cls, op, in_data, axis_chunk_shape, axis_offsets, out_idx
    ):
        raise NotImplementedError

    @classmethod
    def concat_and_pivot(
        cls, op, axis_chunk_shape, out_idx, sorted_chunks, sampled_chunks
    ):
        raise NotImplementedError

    @classmethod
    def partition_local_data(
        cls, op, axis_chunk_shape, sorted_chunks, indices_chunks, concat_pivot_chunk
    ):
        raise NotImplementedError

    @classmethod
    def partition_merge_data(
        cls, op, need_align, return_value, partition_chunks, proxy_chunk
    ):
        raise NotImplementedError

    @classmethod
    def align_partitions_data(
        cls,
        op,
        out_idx,
        in_data,
        partition_sort_chunks,
        partition_indices_chunks,
        sort_info_chunks,
    ):
        raise NotImplementedError


class TensorPSRSOperandMixin(TensorOperandMixin, PSRSOperandMixin):
    @classmethod
    def local_sort_and_regular_sample(
        cls, op, in_data, axis_chunk_shape, axis_offsets, out_idx
    ):
        # stage 1: local sort and regular samples collected
        sorted_chunks, indices_chunks, sampled_chunks = [], [], []
        sampled_dtype = (
            np.dtype([(o, in_data.dtype[o]) for o in op.order])
            if op.order is not None
            else in_data.dtype
        )
        for i in range(axis_chunk_shape):
            idx = list(out_idx)
            idx.insert(op.axis, i)
            in_chunk = in_data.cix[tuple(idx)]
            kind = None if op.psrs_kinds is None else op.psrs_kinds[0]
            chunk_op = PSRSSortRegularSample(
                axis=op.axis,
                order=op.order,
                kind=kind,
                return_indices=op.return_indices,
                n_partition=axis_chunk_shape,
                axis_offset=axis_offsets[i],
                gpu=op.gpu,
            )
            kws = []
            sort_shape = in_chunk.shape
            kws.append(
                {
                    "shape": sort_shape,
                    "order": in_chunk.order,
                    "dtype": in_chunk.dtype,
                    "index": in_chunk.index,
                    "type": "sorted",
                }
            )
            if op.return_indices:
                kws.append(
                    {
                        "shape": sort_shape,
                        "order": TensorOrder.C_ORDER,
                        "dtype": np.dtype(np.int64),
                        "index": in_chunk.index,
                        "type": "argsort",
                    }
                )
            sampled_shape = (axis_chunk_shape,)
            kws.append(
                {
                    "shape": sampled_shape,
                    "order": in_chunk.order,
                    "dtype": sampled_dtype,
                    "index": (i,),
                    "type": "regular_sampled",
                }
            )
            chunks = chunk_op.new_chunks([in_chunk], kws=kws, output_limit=len(kws))
            if len(chunks) == 2:
                sort_chunk, sampled_chunk = chunks
                sorted_chunks.append(sort_chunk)
                sampled_chunks.append(sampled_chunk)
            else:
                sort_chunk, indices_chunk, sampled_chunk = chunks
                sorted_chunks.append(sort_chunk)
                indices_chunks.append(indices_chunk)
                sampled_chunks.append(sampled_chunk)

        return sorted_chunks, indices_chunks, sampled_chunks

    @classmethod
    def concat_and_pivot(
        cls, op, axis_chunk_shape, out_idx, sorted_chunks, sampled_chunks
    ):
        # stage 2: gather and merge samples, choose and broadcast p-1 pivots
        concat_pivot_op = PSRSConcatPivot(
            axis=op.axis,
            order=op.order,
            kind=None if op.psrs_kinds is None else op.psrs_kinds[1],
            dtype=sampled_chunks[0].dtype,
            gpu=op.gpu,
        )
        concat_pivot_shape = (
            sorted_chunks[0].shape[: op.axis]
            + (axis_chunk_shape - 1,)
            + sorted_chunks[0].shape[op.axis + 1 :]
        )
        concat_pivot_index = out_idx[: op.axis] + (0,) + out_idx[op.axis :]
        concat_pivot_chunk = concat_pivot_op.new_chunk(
            sampled_chunks, shape=concat_pivot_shape, index=concat_pivot_index
        )
        return concat_pivot_chunk

    @classmethod
    def partition_local_data(
        cls, op, axis_chunk_shape, sorted_chunks, indices_chunks, concat_pivot_chunk
    ):
        # stage 3: Local data is partitioned
        return_value = op.return_value
        return_indices = op.return_indices
        if return_indices:
            # if return indices and psrs_kind[2] is not None
            # value has to be output
            map_return_value = True
        else:
            map_return_value = return_value
        partition_chunks = []
        length = len(sorted_chunks or indices_chunks)
        for i in range(length):
            chunk_inputs = []
            if sorted_chunks:
                chunk_inputs.append(sorted_chunks[i])
            if indices_chunks:
                chunk_inputs.append(indices_chunks[i])
            chunk_inputs.append(concat_pivot_chunk)
            partition_shuffle_map = PSRSShuffle(
                return_value=map_return_value,
                return_indices=return_indices,
                stage=OperandStage.map,
                axis=op.axis,
                n_partition=axis_chunk_shape,
                input_sorted=op.psrs_kinds[0] is not None,
                order=op.order,
                dtype=chunk_inputs[0].dtype,
                gpu=chunk_inputs[0].op.gpu,
            )
            partition_chunk = partition_shuffle_map.new_chunk(
                chunk_inputs,
                shape=chunk_inputs[0].shape,
                index=chunk_inputs[0].index,
                order=chunk_inputs[0].order,
            )
            partition_chunks.append(partition_chunk)
        return partition_chunks

    @classmethod
    def partition_merge_data(
        cls, op, need_align, return_value, partition_chunks, proxy_chunk
    ):
        # stage 4: all *ith* classes are gathered and merged
        return_value = return_value if return_value is not None else op.return_value
        return_indices = op.return_indices
        partition_sort_chunks, partition_indices_chunks, sort_info_chunks = [], [], []
        for i, partition_chunk in enumerate(partition_chunks):
            kind = None if op.psrs_kinds is None else op.psrs_kinds[2]
            partition_shuffle_reduce = PSRSShuffle(
                return_value=return_value,
                return_indices=return_indices,
                stage=OperandStage.reduce,
                axis=op.axis,
                order=op.order,
                kind=kind,
                reducer_index=(i,),
                n_reducers=len(partition_chunks),
                reducer_ordinal=i,
                dtype=partition_chunk.dtype,
                gpu=partition_chunk.op.gpu,
                need_align=need_align,
            )
            kws = []
            chunk_shape = list(partition_chunk.shape)
            chunk_shape[op.axis] = np.nan
            if return_value:
                kws.append(
                    {
                        "shape": tuple(chunk_shape),
                        "order": partition_chunk.order,
                        "index": partition_chunk.index,
                        "dtype": partition_chunk.dtype,
                        "type": "sorted",
                    }
                )
            if return_indices:
                kws.append(
                    {
                        "shape": tuple(chunk_shape),
                        "order": TensorOrder.C_ORDER,
                        "index": partition_chunk.index,
                        "dtype": np.dtype(np.int64),
                        "type": "argsort",
                    }
                )
            if need_align:
                s = list(chunk_shape)
                s.pop(op.axis)
                kws.append(
                    {
                        "shape": tuple(s),
                        "order": TensorOrder.C_ORDER,
                        "index": partition_chunk.index,
                        "dtype": np.dtype(np.int32),
                        "type": "sort_info",
                    }
                )
            cs = partition_shuffle_reduce.new_chunks([proxy_chunk], kws=kws)
            i = 0
            if return_value:
                partition_sort_chunks.append(cs[0])
                i += 1
            if return_indices:
                partition_indices_chunks.append(cs[i])
            if need_align:
                sort_info_chunks.append(cs[-1])

        return partition_sort_chunks, partition_indices_chunks, sort_info_chunks

    @classmethod
    def align_partitions_data(
        cls,
        op,
        out_idx,
        in_tensor,
        partition_sort_chunks,
        partition_indices_chunks,
        sort_info_chunks,
    ):
        return_value, return_indices = op.return_value, op.return_indices
        align_map_chunks = []
        length = len(partition_sort_chunks or partition_indices_chunks)
        for i in range(length):
            chunk_inputs = []
            if return_value:
                chunk_inputs.append(partition_sort_chunks[i])
            if return_indices:
                chunk_inputs.append(partition_indices_chunks[i])
            chunk_inputs.extend(sort_info_chunks)
            align_map_op = PSRSAlign(
                return_value=return_value,
                return_indices=return_indices,
                stage=OperandStage.map,
                axis=op.axis,
                output_sizes=list(in_tensor.nsplits[op.axis]),
                dtype=chunk_inputs[0].dtype,
                gpu=chunk_inputs[0].op.gpu,
            )
            align_map_chunk = align_map_op.new_chunk(
                chunk_inputs,
                shape=chunk_inputs[0].shape,
                index=chunk_inputs[0].index,
                order=TensorOrder.C_ORDER,
            )
            align_map_chunks.append(align_map_chunk)
        proxy_chunk = TensorShuffleProxy(dtype=align_map_chunks[0].dtype).new_chunk(
            align_map_chunks, shape=()
        )
        align_reduce_value_chunks, align_reduce_indices_chunks = [], []
        for i, align_map_chunk in enumerate(align_map_chunks):
            align_reduce_op = PSRSAlign(
                return_value=return_value,
                return_indices=return_indices,
                stage=OperandStage.reduce,
                axis=op.axis,
                reducer_index=(i,),
                reducer_ordinal=i,
                n_reducers=len(align_map_chunks),
                dtype=align_map_chunk.dtype,
                gpu=align_map_chunk.op.gpu,
            )
            idx = list(out_idx)
            idx.insert(op.axis, i)
            in_chunk = in_tensor.cix[tuple(idx)]
            kws = []
            if return_value:
                kws.append(
                    {
                        "shape": in_chunk.shape,
                        "index": in_chunk.index,
                        "order": in_chunk.order,
                        "dtype": in_chunk.dtype,
                        "type": "sorted",
                    }
                )
            if return_indices:
                kws.append(
                    {
                        "shape": in_chunk.shape,
                        "index": in_chunk.index,
                        "order": TensorOrder.C_ORDER,
                        "dtype": np.dtype(np.int64),
                        "type": "argsort",
                    }
                )
            align_reduce_chunks = align_reduce_op.new_chunks([proxy_chunk], kws=kws)
            if return_value:
                align_reduce_value_chunks.append(align_reduce_chunks[0])
            if return_indices:
                align_reduce_indices_chunks.append(align_reduce_chunks[-1])

        return align_reduce_value_chunks, align_reduce_indices_chunks


def _sort(a, op, xp, axis=None, kind=None, order=None, inplace=False):
    axis = axis if axis is not None else op.axis
    kind = kind if kind is not None else op.kind
    order = order if order is not None else op.order
    if xp is np:
        method = a.sort if inplace else partial(np.sort, a)
        return method(axis=axis, kind=kind, order=order)
    else:  # pragma: no cover
        # cupy does not support structure type
        assert xp is cp
        assert order is not None
        method = a.sort if inplace else partial(cp.sort, a)
        # cupy does not support kind, thus just ignore it
        return method(axis=axis)


def _argsort(a, op, xp, axis=None, kind=None, order=None):
    axis = axis if axis is not None else op.axis
    kind = kind if kind is not None else op.kind
    order = order if order is not None else op.order
    if xp is np:
        return np.argsort(a, axis=axis, kind=kind, order=order)
    else:  # pragma: no cover
        # cupy does not support structure type
        assert xp is cp
        assert order is not None
        return cp.argsort(a, axis=axis)


class PSRSSortRegularSample(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.PSRS_SORT_REGULAR_SMAPLE

    _axis = Int32Field("axis")
    _order = ListField("order", FieldTypes.string)
    _kind = StringField("kind")
    _return_indices = BoolField("return_indices")
    _n_partition = Int32Field("n_partition")
    _axis_offset = AnyField("axis_offset")

    def __init__(
        self,
        axis=None,
        order=None,
        kind=None,
        return_indices=None,
        n_partition=None,
        axis_offset=None,
        dtype=None,
        gpu=None,
        **kw
    ):
        super().__init__(
            _axis=axis,
            _order=order,
            _kind=kind,
            _return_indices=return_indices,
            _n_partition=n_partition,
            _axis_offset=axis_offset,
            dtype=dtype,
            gpu=gpu,
            **kw
        )

    @property
    def axis(self):
        return self._axis

    @property
    def order(self):
        return self._order

    @property
    def kind(self):
        return self._kind

    @property
    def return_indices(self):
        return self._return_indices

    @property
    def n_partition(self):
        return self._n_partition

    @property
    def axis_offset(self):
        return self._axis_offset

    @property
    def output_limit(self):
        # return sorted tensor, indices(optional) and regular sampled tensor
        return 2 if not self._return_indices else 3

    @classmethod
    def execute(cls, ctx, op):
        (a,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        if len(a) == 0:
            # when chunk is empty, return the empty chunk itself
            ctx[op.outputs[0].key] = ctx[op.outputs[-1].key] = a
            return

        with device(device_id):
            n = op.n_partition
            w = a.shape[op.axis] * 1.0 / (n + 1)
            if not op.return_indices:
                if op.kind is not None:
                    # sort
                    res = ctx[op.outputs[0].key] = _sort(a, op, xp)
                else:
                    # do not sort, prepare for sample by `xp.partition`
                    kth = xp.linspace(
                        max(w - 1, 0), a.shape[op.axis] - 1, num=n, endpoint=False
                    ).astype(int)
                    ctx[op.outputs[0].key] = res = xp.partition(
                        a, kth, axis=op.axis, order=op.order
                    )
            else:
                if op.kind is not None:
                    # argsort
                    indices = _argsort(a, op, xp)
                else:
                    # do not sort, use `xp.argpartition`
                    kth = xp.linspace(
                        max(w - 1, 0), a.shape[op.axis] - 1, num=n, endpoint=False
                    ).astype(int)
                    indices = xp.argpartition(a, kth, axis=op.axis, order=op.order)
                ctx[op.outputs[0].key] = res = xp.take_along_axis(a, indices, op.axis)
                ctx[op.outputs[1].key] = op.axis_offset + indices

            # do regular sample
            if op.order is not None:
                res = res[op.order]
            slc = xp.linspace(
                max(w - 1, 0), a.shape[op.axis] - 1, num=n, endpoint=False
            ).astype(int)
            slc = (slice(None),) * op.axis + (slc,)
            ctx[op.outputs[-1].key] = res[slc]


class PSRSConcatPivot(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.PSRS_CONCAT_PIVOT

    _axis = Int32Field("axis")
    _order = ListField("order", FieldTypes.string)
    _kind = StringField("kind")

    def __init__(self, axis=None, order=None, kind=None, dtype=None, gpu=None, **kw):
        super().__init__(
            _axis=axis, _order=order, _kind=kind, dtype=dtype, gpu=gpu, **kw
        )

    @property
    def axis(self):
        return self._axis

    @property
    def order(self):
        return self._order

    @property
    def kind(self):
        return self._kind

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs if len(ctx[c.key]) > 0],
            device=op.device,
            ret_extra=True,
        )

        with device(device_id):
            a = xp.concatenate(inputs, axis=op.axis)
            p = len(inputs)
            assert a.shape[op.axis] == p * len(op.inputs)

            if op.kind is not None:
                # sort
                _sort(a, op, xp, inplace=True)
            else:
                # prepare for sampling via `partition`
                kth = xp.linspace(
                    p - 1, a.shape[op.axis] - 1, num=p - 1, endpoint=False
                ).astype(int)
                a.partition(kth, axis=op.axis)

            select = xp.linspace(
                p - 1, a.shape[op.axis] - 1, num=len(op.inputs) - 1, endpoint=False
            ).astype(int)
            slc = (slice(None),) * op.axis + (select,)
            ctx[op.outputs[0].key] = a[slc]


class PSRSShuffle(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.PSRS_SHUFFLE

    # public
    _return_value = BoolField("return_value")
    _return_indices = BoolField("return_indices")

    # for shuffle map
    _axis = Int32Field("axis")
    _order = ListField("order", FieldTypes.string)
    _n_partition = Int32Field("n_partition")
    _input_sorted = BoolField("input_sorted")

    # for shuffle reduce
    _kind = StringField("kind")
    _need_align = BoolField("need_align")

    def __init__(
        self,
        return_value=None,
        return_indices=None,
        axis=None,
        order=None,
        n_partition=None,
        input_sorted=None,
        kind=None,
        need_align=None,
        **kw
    ):
        super().__init__(
            _return_value=return_value,
            _return_indices=return_indices,
            _axis=axis,
            _order=order,
            _n_partition=n_partition,
            _input_sorted=input_sorted,
            _kind=kind,
            _need_align=need_align,
            **kw
        )

    @property
    def return_value(self):
        return self._return_value

    @property
    def return_indices(self):
        return self._return_indices

    @property
    def axis(self):
        return self._axis

    @property
    def order(self):
        return self._order

    @property
    def n_partition(self):
        return self._n_partition

    @property
    def input_sorted(self):
        return self._input_sorted

    @property
    def kind(self):
        return self._kind

    @property
    def need_align(self):
        return self._need_align

    @property
    def output_limit(self):
        if self.stage == OperandStage.map:
            return 1
        else:
            limit = int(bool(self._return_value)) + int(bool(self._return_indices))
            if self._need_align:
                limit += 1
            return limit

    @classmethod
    def _execute_map(cls, ctx, op):
        return_value = op.return_value
        return_indices = op.return_indices

        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )
        out = op.outputs[0]
        a = inputs[0]
        pivots = inputs[-1]
        a_indices = None
        if return_indices:
            a_indices = inputs[-2]

        with device(device_id):
            shape = tuple(s for i, s in enumerate(a.shape) if i != op.axis)
            reduce_outputs = [
                np.empty(shape, dtype=object) for _ in range(op.n_partition)
            ]
            for idx in itertools.product(*(range(s) for s in shape)):
                slc = list(idx)
                slc.insert(op.axis, slice(None))
                slc = tuple(slc)
                a_1d, pivots_1d = a[slc], pivots[slc]
                a_indices_1d = a_indices[slc] if a_indices is not None else None
                raw_a_1d = a_1d
                if op.order is not None:
                    a_1d = a_1d[op.order]
                if op.input_sorted:
                    # a is sorted already
                    poses = xp.searchsorted(a_1d, pivots_1d, side="right")
                    poses = (None,) + tuple(poses) + (None,)
                    for i in range(op.n_partition):
                        reduce_out = []
                        if return_value:
                            values = raw_a_1d[poses[i] : poses[i + 1]]
                            reduce_out.append(values)
                        if return_indices:
                            indices = a_indices_1d[poses[i] : poses[i + 1]]
                            reduce_out.append(indices)
                        reduce_outputs[i][idx] = tuple(reduce_out)
                else:
                    # a is not sorted, search every element in pivots
                    out_idxes = xp.searchsorted(pivots_1d, a_1d, side="right")
                    for i in range(op.n_partition):
                        cond = out_idxes == i
                        reduce_out = []
                        if return_value:
                            values = raw_a_1d[cond]
                            reduce_out.append(values)
                        if return_indices:
                            indices = a_indices_1d[cond]
                            reduce_out.append(indices)
                        reduce_outputs[i][idx] = tuple(reduce_out)
            for i in range(op.n_partition):
                ctx[out.key, (i,)] = tuple(reduce_outputs[i].ravel())

    @classmethod
    def _execute_reduce(cls, ctx, op: "PSRSShuffle"):
        raw_inputs = list(op.iter_mapper_data(ctx))
        # flatten inputs
        flatten_inputs = flatten(raw_inputs)
        inputs, device_id, xp = as_same_device(
            flatten_inputs, device=op.device, ret_extra=True
        )
        # organize back inputs
        inputs = stack_back(inputs, raw_inputs)

        out = op.outputs[0]
        extra_shape = list(out.shape)
        extra_shape.pop(op.axis)

        return_value = op.return_value
        return_indices = op.return_indices

        with device(device_id):
            sort_res = np.empty(len(inputs[0]), dtype=object)
            if extra_shape:
                sort_res = sort_res.reshape(*extra_shape)
            sort_info = np.empty(sort_res.shape, dtype=np.int32)
            it = itertools.count(0)
            for inps in zip(*inputs):
                cur = itertools.count()
                values, indices = None, None
                ret = []
                if return_value or len(inps[0]) == 2:
                    i = next(cur)
                    values = xp.concatenate([inp[i] for inp in inps])
                    if return_value:
                        ret.append(values)
                if return_indices:
                    i = next(cur)
                    indices = xp.concatenate([inp[i] for inp in inps])
                    ret.append(indices)

                if op.kind is not None:
                    # sort only if kind specified
                    if return_indices:
                        # if kind specified and return_indices
                        # values cannot be None
                        assert values is not None
                        values_indices = _argsort(values, op, xp, axis=0)
                        if return_value:
                            xp.take(values, values_indices, out=values)
                        xp.take(indices, values_indices, out=indices)
                    else:
                        _sort(values, op, xp, axis=0, inplace=True)

                j = next(it)
                sort_res.ravel()[j] = ret
                sort_info.ravel()[j] = len(ret[0])

            if not op.need_align:
                assert len(sort_res) == 1
                shape = list(extra_shape)
                shape.insert(op.axis, len(sort_res[0]))
                i = 0
                if return_value:
                    ctx[op.outputs[0].key] = sort_res[0][i]
                    i += 1
                if return_indices:
                    ctx[op.outputs[i].key] = sort_res[0][i]
            else:
                i = 0
                if return_value:
                    ctx[op.outputs[0].key] = tuple(r[0] for r in sort_res.ravel())
                    i += 1
                if return_indices:
                    ctx[op.outputs[i].key] = tuple(r[i] for r in sort_res.ravel())
                ctx[op.outputs[-1].key] = sort_info

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)


class PSRSAlign(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.PSRS_ALIGN

    _return_value = BoolField("return_value")
    _return_indices = BoolField("return_indices")
    _axis = Int32Field("axis")
    _output_sizes = ListField("output_sizes", FieldTypes.int32)

    def __init__(
        self, return_value=None, return_indices=None, axis=None, output_sizes=None, **kw
    ):
        super().__init__(
            _return_value=return_value,
            _return_indices=return_indices,
            _axis=axis,
            _output_sizes=output_sizes,
            **kw
        )

    @property
    def return_value(self):
        return self._return_value

    @property
    def return_indices(self):
        return self._return_indices

    @property
    def axis(self):
        return self._axis

    @property
    def output_sizes(self):
        return self._output_sizes

    @property
    def output_limit(self):
        if self.stage == OperandStage.map:
            return 1
        else:
            return int(bool(self._return_value)) + int(bool(self._return_indices))

    @classmethod
    def _execute_map(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )
        sort_res, sort_indices = None, None
        i = 0
        if op.return_value:
            sort_res = inputs[0]
            i += 1
        if op.return_indices:
            sort_indices = inputs[i]
            i += 1
        sort_infos = inputs[i:]
        out = op.outputs[0]

        with device(device_id):
            length = len(sort_res or sort_indices)
            outs = np.empty((len(op.output_sizes), length), dtype=object)
            out_sizes = op.output_sizes
            cum_out_sizes = (0,) + tuple(np.cumsum(out_sizes))
            for i in range(length):
                sort_1d = sort_res[i] if sort_res is not None else None
                indices_1d = sort_indices[i] if sort_indices is not None else None
                sort_lengths = [sort_info.flat[i] for sort_info in sort_infos]
                cum_sort_lengths = (0,) + tuple(np.cumsum(sort_lengths))
                j = out.index[op.axis]
                start_pos = cum_sort_lengths[j]
                end_pos = cum_sort_lengths[j + 1]
                out_idx_start, out_idx_end = np.searchsorted(
                    cum_out_sizes, [start_pos, end_pos]
                )
                out_idx_start = max(out_idx_start - 1, 0)
                for out_idx in range(out_idx_start, out_idx_end):
                    out_start_pos = cum_out_sizes[out_idx]
                    out_end_pos = cum_out_sizes[out_idx + 1]
                    s = max(start_pos, out_start_pos)
                    size = max(min(end_pos, out_end_pos) - s, 0)
                    s = max(0, s - start_pos)
                    ret = []
                    if sort_1d is not None:
                        ret.append(sort_1d[s : s + size])
                    if indices_1d is not None:
                        ret.append(indices_1d[s : s + size])
                    outs[out_idx, i] = tuple(ret)

            for idx in range(len(op.output_sizes)):
                ret = []
                for ar in outs[idx]:
                    if ar is None:
                        item = []
                        if sort_res is not None:
                            item.append(xp.empty((0,), dtype=out.dtype))
                        if sort_indices is not None:
                            item.append(xp.empty((0,), dtype=np.dtype(np.int64)))
                        ret.append(tuple(item))
                    else:
                        ret.append(ar)
                ctx[op.outputs[0].key, (idx,)] = tuple(ret)

    @classmethod
    def _execute_reduce(cls, ctx, op: "PSRSAlign"):
        axis = op.axis
        raw_inputs = list(op.iter_mapper_data(ctx))
        flatten_inputs = flatten(raw_inputs)
        inputs, device_id, xp = as_same_device(
            flatten_inputs, device=op.device, ret_extra=True
        )
        inputs = stack_back(flatten_inputs, raw_inputs)

        out = op.outputs[0]
        extra_shape = list(out.shape)
        extra_shape.pop(axis)

        return_value = op.return_value
        return_indices = op.return_indices

        with device(device_id):
            if return_value:
                values_res = xp.empty(out.shape, dtype=out.dtype)
            else:
                values_res = None
            if return_indices:
                indices_res = xp.empty(out.shape, dtype=np.dtype(np.int64))
            else:
                indices_res = None
            it = itertools.product(
                *(range(s) for i, s in enumerate(out.shape) if i != axis)
            )
            for inps in zip(*inputs):
                slc = list(next(it))
                slc.insert(op.axis, slice(None))
                i = 0
                if return_value:
                    value_concat_1d = xp.concatenate([inp[0] for inp in inps])
                    values_res[tuple(slc)] = value_concat_1d
                    i += 1
                if return_indices:
                    ind_concat_id = xp.concatenate([inp[i] for inp in inps])
                    indices_res[tuple(slc)] = ind_concat_id

            i = 0
            if return_value:
                ctx[op.outputs[0].key] = values_res.astype(
                    values_res.dtype, order=op.outputs[0].order.value
                )
                i += 1
            if return_indices:
                ctx[op.outputs[i].key] = indices_res.astype(
                    indices_res.dtype, order=op.outputs[i].order.value
                )

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import recursive_tile
from ...serialization.serializables import KeyField, AnyField, BoolField
from ...utils import has_unknown_shape
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_df, parse_index


class DataFrameStack(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.STACK

    _input_df = KeyField("input_df")
    _level = AnyField("level")
    _dropna = BoolField("dropna")

    def __init__(self, input_df=None, level=None, dropna=None, **kw):
        super().__init__(_input_df=input_df, _level=level, _dropna=dropna, **kw)

    @property
    def input_df(self):
        return self._input_df

    @property
    def level(self):
        return self._level

    @property
    def dropna(self):
        return self._dropna

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input_df = self._inputs[0]

    @classmethod
    def _calc_size(cls, size: int, level: Union[List, int], dtypes: pd.Series):
        index = dtypes.index

        if not isinstance(index, pd.MultiIndex):
            return size * len(index)

        if isinstance(level, int):
            level = [level]
        return size * np.prod([index.levshape[lev] for lev in level]).item()

    def __call__(self, input_df):
        test_df = build_df(input_df)
        test_df = test_df.stack(level=self._level, dropna=self._dropna)
        if self._dropna:
            size = np.nan
        else:
            size = self._calc_size(input_df.shape[0], self._level, input_df.dtypes)
        if test_df.ndim == 1:
            shape = (size,)
            return self.new_series(
                [input_df],
                shape=shape,
                dtype=test_df.dtype,
                index_value=parse_index(test_df.index, input_df),
                name=test_df.name,
            )
        else:
            shape = (size, test_df.shape[1])
            return self.new_dataframe(
                [input_df],
                shape=shape,
                dtypes=test_df.dtypes,
                index_value=parse_index(test_df.index, input_df),
                columns_value=parse_index(test_df.columns, store_data=True),
            )

    @classmethod
    def tile(cls, op: "DataFrameStack"):
        input_df = op.input_df
        out = op.outputs[0]
        out_index = out.index_value.to_pandas()

        if input_df.chunk_shape[1] > 1:
            # rechunk into 1 chunk on axis 1
            if has_unknown_shape(input_df):
                yield
            input_df = yield from recursive_tile(
                input_df.rechunk({1: input_df.shape[1]})
            )

        out_chunks = []
        for c in input_df.chunks:
            chunk_op = op.copy().reset_key()
            if op.dropna:
                size = np.nan
            else:
                size = cls._calc_size(c.shape[0], op.level, c.dtypes)
            if out.ndim == 1:
                kw = {
                    "shape": (size,),
                    "index": (c.index[0],),
                    "dtype": out.dtype,
                    "index_value": parse_index(out_index, c),
                    "name": out.name,
                }
            else:
                kw = {
                    "shape": (size, out.shape[1]),
                    "index": (c.index[0], 0),
                    "dtypes": out.dtypes,
                    "index_value": parse_index(out_index, c),
                    "columns_value": out.columns_value,
                }
            out_chunk = chunk_op.new_chunk([c], **kw)
            out_chunks.append(out_chunk)

        params = out.params
        if out.ndim == 1:
            params["nsplits"] = (tuple(out_c.shape[0] for out_c in out_chunks),)
        else:
            params["nsplits"] = (
                tuple(out_c.shape[0] for out_c in out_chunks),
                (out.shape[1],),
            )
        params["chunks"] = out_chunks
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op: "DataFrameStack"):
        inp: pd.DataFrame = ctx[op.input_df.key]
        ctx[op.outputs[0].key] = inp.stack(level=op.level, dropna=op.dropna)


def stack(df, level=-1, dropna=True):
    """
    Stack the prescribed level(s) from columns to index.

    Return a reshaped DataFrame or Series having a multi-level
    index with one or more new inner-most levels compared to the current
    DataFrame. The new inner-most levels are created by pivoting the
    columns of the current dataframe:

      - if the columns have a single level, the output is a Series;
      - if the columns have multiple levels, the new index
        level(s) is (are) taken from the prescribed level(s) and
        the output is a DataFrame.

    Parameters
    ----------
    level : int, str, list, default -1
        Level(s) to stack from the column axis onto the index
        axis, defined as one index or label, or a list of indices
        or labels.
    dropna : bool, default True
        Whether to drop rows in the resulting Frame/Series with
        missing values. Stacking a column level onto the index
        axis can create combinations of index and column values
        that are missing from the original dataframe. See Examples
        section.

    Returns
    -------
    DataFrame or Series
        Stacked dataframe or series.

    See Also
    --------
    DataFrame.unstack : Unstack prescribed level(s) from index axis
         onto column axis.
    DataFrame.pivot : Reshape dataframe from long format to wide
         format.
    DataFrame.pivot_table : Create a spreadsheet-style pivot table
         as a DataFrame.

    Notes
    -----
    The function is named by analogy with a collection of books
    being reorganized from being side by side on a horizontal
    position (the columns of the dataframe) to being stacked
    vertically on top of each other (in the index of the
    dataframe).

    Examples
    --------
    **Single level columns**

    >>> import mars.dataframe as md
    >>> df_single_level_cols = md.DataFrame([[0, 1], [2, 3]],
    ...                                     index=['cat', 'dog'],
    ...                                     columns=['weight', 'height'])

    Stacking a dataframe with a single level column axis returns a Series:

    >>> df_single_level_cols.execute()
         weight height
    cat       0      1
    dog       2      3
    >>> df_single_level_cols.stack().execute()
    cat  weight    0
         height    1
    dog  weight    2
         height    3
    dtype: int64

    **Multi level columns: simple case**

    >>> multicol1 = pd.MultiIndex.from_tuples([('weight', 'kg'),
    ...                                        ('weight', 'pounds')])
    >>> df_multi_level_cols1 = md.DataFrame([[1, 2], [2, 4]],
    ...                                     index=['cat', 'dog'],
    ...                                     columns=multicol1)

    Stacking a dataframe with a multi-level column axis:

    >>> df_multi_level_cols1.execute()
         weight
             kg    pounds
    cat       1        2
    dog       2        4
    >>> df_multi_level_cols1.stack().execute()
                weight
    cat kg           1
        pounds       2
    dog kg           2
        pounds       4

    **Missing values**

    >>> multicol2 = pd.MultiIndex.from_tuples([('weight', 'kg'),
    ...                                        ('height', 'm')])
    >>> df_multi_level_cols2 = md.DataFrame([[1.0, 2.0], [3.0, 4.0]],
    ...                                     index=['cat', 'dog'],
    ...                                     columns=multicol2)

    It is common to have missing values when stacking a dataframe
    with multi-level columns, as the stacked dataframe typically
    has more values than the original dataframe. Missing values
    are filled with NaNs:

    >>> df_multi_level_cols2.execute()
        weight height
            kg      m
    cat    1.0    2.0
    dog    3.0    4.0
    >>> df_multi_level_cols2.stack().execute()
            height  weight
    cat kg     NaN     1.0
        m      2.0     NaN
    dog kg     NaN     3.0
        m      4.0     NaN

    **Prescribing the level(s) to be stacked**

    The first parameter controls which level or levels are stacked:

    >>> df_multi_level_cols2.stack(0).execute()
                 kg    m
    cat height  NaN  2.0
        weight  1.0  NaN
    dog height  NaN  4.0
        weight  3.0  NaN
    >>> df_multi_level_cols2.stack([0, 1]).execute()
    cat  height  m     2.0
         weight  kg    1.0
    dog  height  m     4.0
         weight  kg    3.0
    dtype: float64

    **Dropping missing values**

    >>> df_multi_level_cols3 = md.DataFrame([[None, 1.0], [2.0, 3.0]],
    ...                                     index=['cat', 'dog'],
    ...                                     columns=multicol2)

    Note that rows where all values are missing are dropped by
    default but this behaviour can be controlled via the dropna
    keyword parameter:

    >>> df_multi_level_cols3.execute()
        weight height
            kg      m
    cat    NaN    1.0
    dog    2.0    3.0
    >>> df_multi_level_cols3.stack(dropna=False).execute()
            height  weight
    cat kg     NaN     NaN
        m      1.0     NaN
    dog kg     NaN     2.0
        m      3.0     NaN
    >>> df_multi_level_cols3.stack(dropna=True).execute()
            height  weight
    cat m      1.0     NaN
    dog kg     NaN     2.0
        m      3.0     NaN
    """
    op = DataFrameStack(input_df=df, level=level, dropna=dropna)
    return op(df)



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
from typing import AsyncGenerator, Dict, List, Optional, TypeVar

from ...services.cluster import WebClusterAPI
from ...services.cluster.backends import (
    register_cluster_backend,
    AbstractClusterBackend,
)
from ...services.cluster.core import NodeRole
from ..utils import wait_all_supervisors_ready, next_in_thread
from .config import MarsReplicationConfig

logger = logging.getLogger(__name__)
RetType = TypeVar("RetType")


@register_cluster_backend
class K8SClusterBackend(AbstractClusterBackend):
    name = "k8s"

    def __init__(
        self, node_role=None, pool_address=None, k8s_config=None, k8s_namespace=None
    ):
        from kubernetes import client

        self._node_role = node_role
        self._pool_address = pool_address
        self._k8s_config = k8s_config

        verify_ssl = bool(int(os.environ.get("KUBE_VERIFY_SSL", "1")))
        if not verify_ssl:
            c = client.Configuration()
            c.verify_ssl = False
            client.Configuration.set_default(c)

        self._k8s_namespace = (
            k8s_namespace or os.environ.get("MARS_K8S_POD_NAMESPACE") or "default"
        )
        self._service_name = os.environ.get("MARS_K8S_SERVICE_NAME")
        self._full_label_selector = None
        self._client = client.CoreV1Api(client.ApiClient(self._k8s_config))

        self._pod_to_ep = dict()

    @classmethod
    async def create(
        cls, node_role: NodeRole, lookup_address: Optional[str], pool_address: str
    ) -> "AbstractClusterBackend":
        from kubernetes import config, client

        if lookup_address is None:
            k8s_namespace = None
            k8s_config = config.load_incluster_config()
        else:
            address_parts = lookup_address.rsplit("?", 1)
            k8s_namespace = None if len(address_parts) == 1 else address_parts[1]

            k8s_config = client.Configuration()
            if "://" in address_parts[0]:
                k8s_config.host = address_parts[0]
            else:
                config.load_kube_config(
                    address_parts[0], client_configuration=k8s_config
                )
        return cls(node_role, pool_address, k8s_config, k8s_namespace)

    def __reduce__(self):
        return (
            type(self),
            (
                self._node_role,
                self._pool_address,
                self._k8s_config,
                self._k8s_namespace,
            ),
        )

    @staticmethod
    def _format_endpoint_query_result(result: Dict, filter_ready: bool = True):
        port = os.environ["MARS_K8S_SERVICE_PORT"]
        endpoints = [
            f"{addr['ip']}:{port}" for addr in result["subsets"][0]["addresses"] or []
        ]
        if not filter_ready:
            endpoints = [
                f"{addr['ip']}:{port}"
                for addr in result["subsets"][0]["not_ready_addresses"] or []
            ]
        return endpoints

    def _get_web_cluster_api(self):
        supervisor_web_port = os.environ["MARS_K8S_SUPERVISOR_WEB_PORT"]
        web_url = (
            f"http://{self._service_name}.{self._k8s_namespace}:{supervisor_web_port}"
        )
        api = WebClusterAPI(web_url)
        return api

    async def _watch_supervisors_by_service_api(
        self,
    ) -> AsyncGenerator[List[str], None]:
        from urllib3.exceptions import ReadTimeoutError
        from kubernetes.watch import Watch as K8SWatch

        w = K8SWatch()

        while True:
            streamer = w.stream(
                self._client.list_namespaced_endpoints,
                namespace=self._k8s_namespace,
                label_selector=f"mars/service-name={self._service_name}",
                timeout_seconds=60,
            )
            while True:
                try:
                    event = await next_in_thread(streamer)
                    obj_dict = event["object"].to_dict()
                    yield self._format_endpoint_query_result(obj_dict)
                except (ReadTimeoutError, StopAsyncIteration):
                    break
                except:  # noqa: E722  # pragma: no cover  # pylint: disable=bare-except
                    logger.exception("Unexpected error when watching on kubernetes")
                    break

    async def _watch_supervisors_by_cluster_web_api(self):
        while True:
            try:
                api = self._get_web_cluster_api()
                async for supervisors in api.watch_supervisors():
                    yield supervisors
            except (OSError, asyncio.TimeoutError):
                pass

    async def _get_supervisors_by_service_api(
        self, filter_ready: bool = True
    ) -> List[str]:
        result = (
            await asyncio.to_thread(
                self._client.read_namespaced_endpoints,
                name=self._service_name,
                namespace=self._k8s_namespace,
            )
        ).to_dict()
        return self._format_endpoint_query_result(result, filter_ready=filter_ready)

    async def _get_supervisors_by_cluster_web_api(self, filter_ready: bool = True):
        api = self._get_web_cluster_api()
        try:
            supervisors = await api.get_supervisors(filter_ready=filter_ready)
            return supervisors
        except (OSError, asyncio.TimeoutError):  # pragma: no cover
            return []

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        if self._node_role == NodeRole.SUPERVISOR:
            return await self._get_supervisors_by_service_api(filter_ready)
        else:
            return await self._get_supervisors_by_cluster_web_api(filter_ready)

    async def watch_supervisors(self) -> AsyncGenerator[List[str], None]:
        if self._node_role == NodeRole.SUPERVISOR:
            watch_fun = self._watch_supervisors_by_service_api
        else:
            watch_fun = self._watch_supervisors_by_cluster_web_api

        try:
            async for supervisors in watch_fun():
                yield supervisors
        except asyncio.CancelledError:
            pass

    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> str:
        raise NotImplementedError

    async def release_worker(self, address: str):
        raise NotImplementedError

    async def reconstruct_worker(self, address: str):
        raise NotImplementedError


class K8SServiceMixin:
    @staticmethod
    def write_pid_file():
        with open("/tmp/mars-service.pid", "w") as pid_file:
            pid_file.write(str(os.getpid()))

    async def wait_all_supervisors_ready(self):
        """
        Wait till all containers are ready
        """
        await wait_all_supervisors_ready(self.args.endpoint)

    async def start_readiness_server(self):
        readiness_port = os.environ.get(
            "MARS_K8S_READINESS_PORT", MarsReplicationConfig.default_readiness_port
        )
        self._readiness_server = await asyncio.start_server(
            lambda r, w: None, port=readiness_port
        )

    async def stop_readiness_server(self):
        self._readiness_server.close()
        await self._readiness_server.wait_closed()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Dict

from ....utils import lazy_import, lazy_import_on_load
from ..metric import (
    AbstractMetric,
    AbstractCounter,
    AbstractGauge,
    AbstractMeter,
    AbstractHistogram,
)

ray_metrics = lazy_import("ray.util.metrics", rename="ray_metrics")

_ray_gauge_set_available = None


@lazy_import_on_load(ray_metrics)
def _reload_ray_gauge_set_available():
    """
    Note: Gauge `record` method is deprecated in ray 1.3.0 version, so here
    make it compatible with the old and new ray versions.
    """
    global _ray_gauge_set_available

    if _ray_gauge_set_available is not None:
        return _ray_gauge_set_available
    _ray_gauge_set_available = (
        True if ray_metrics and hasattr(ray_metrics.Gauge, "set") else False
    )
    return _ray_gauge_set_available


class RayMetricMixin(AbstractMetric):
    def _init(self):
        _reload_ray_gauge_set_available()

        if ray_metrics is not None:  # pragma: no branch
            self._metric = ray_metrics.Gauge(
                self._name, self._description, self._tag_keys
            )

    def _record(self, value=1, tags: Optional[Dict[str, str]] = None):
        if _ray_gauge_set_available:
            self._metric.set(value, tags)
        elif ray_metrics is not None:  # pragma: no branch
            self._metric.record(value, tags)


class Counter(RayMetricMixin, AbstractCounter):
    pass


class Gauge(RayMetricMixin, AbstractGauge):
    pass


class Meter(RayMetricMixin, AbstractMeter):
    pass


class Histogram(RayMetricMixin, AbstractHistogram):
    pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ... import opcodes
from ...core import recursive_tile
from ...serialization.serializables import AnyField, Int8Field, Int64Field
from ..core import DATAFRAME_TYPE, OutputType
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_df, build_empty_series, validate_axis
from .shift import DataFrameShift


class DataFrameDiff(DataFrameOperandMixin, DataFrameOperand):
    _op_type_ = opcodes.DIFF

    _periods = Int64Field("periods")
    _axis = Int8Field("axis")

    _bool_columns = AnyField("bool_columns")

    @property
    def periods(self):
        return self._periods

    @property
    def axis(self):
        return self._axis

    @property
    def bool_columns(self):
        return self._bool_columns

    def __init__(self, periods=None, axis=None, bool_columns=None, **kw):
        super().__init__(_periods=periods, _axis=axis, _bool_columns=bool_columns, **kw)

    def __call__(self, df_or_series):
        params = df_or_series.params.copy()

        if isinstance(df_or_series, DATAFRAME_TYPE):
            self.output_types = [OutputType.dataframe]
            mock_obj = build_empty_df(df_or_series.dtypes)
            params["dtypes"] = mock_obj.diff().dtypes
        else:
            self.output_types = [OutputType.series]
            mock_obj = build_empty_series(df_or_series.dtype, name=df_or_series.name)
            params["dtype"] = mock_obj.diff().dtype

        return self.new_tileable([df_or_series], **params)

    @classmethod
    def tile(cls, op):
        in_obj = op.inputs[0]
        out_obj = op.outputs[0]
        axis = op.axis or 0

        if in_obj.chunk_shape[axis] > 1:
            shifted = yield from recursive_tile(
                DataFrameShift(periods=op.periods, axis=axis)(in_obj)
            )
            shift_chunks = shifted.chunks
        else:
            shift_chunks = itertools.repeat(None)

        chunks = []
        bool_columns_dict = dict()
        for in_chunk, shift_chunk in zip(in_obj.chunks, shift_chunks):
            params = in_chunk.params.copy()
            if in_chunk.ndim == 2:
                params["dtypes"] = out_obj.dtypes[in_chunk.dtypes.index]
                try:
                    bool_columns = bool_columns_dict[in_chunk.index[1]]
                except KeyError:
                    bool_columns = bool_columns_dict[in_chunk.index[1]] = [
                        col
                        for col, dt in in_chunk.dtypes.items()
                        if dt == np.dtype(bool)
                    ]
            else:
                params["dtype"] = out_obj.dtype
                bool_columns = in_chunk.dtype == np.dtype(bool)

            new_op = op.copy().reset_key()
            new_op._bool_columns = bool_columns

            if shift_chunk is None:
                chunks.append(new_op.new_chunk([in_chunk], **params))
            else:
                chunks.append(new_op.new_chunk([in_chunk, shift_chunk], **params))

        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            [in_obj], chunks=chunks, nsplits=in_obj.nsplits, **out_obj.params
        )

    @classmethod
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        if len(op.inputs) == 1:
            if in_data.ndim == 2:
                try:
                    ctx[op.outputs[0].key] = in_data.diff(
                        periods=op.periods, axis=op.axis
                    )
                except ValueError:
                    ctx[op.outputs[0].key] = in_data.copy().diff(
                        periods=op.periods, axis=op.axis
                    )
            else:
                ctx[op.outputs[0].key] = in_data.diff(periods=op.periods)
        else:
            in_shift = ctx[op.inputs[1].key]
            result = in_data - in_shift
            if op.bool_columns:
                if in_data.ndim == 2:
                    result.replace(
                        {c: {1: True, -1: True, 0: False} for c in op.bool_columns},
                        inplace=True,
                    )
                else:
                    result.replace({1: True, -1: True, 0: False}, inplace=True)
            ctx[op.outputs[0].key] = result


def df_diff(df, periods=1, axis=0):
    """
    First discrete difference of element.
    Calculates the difference of a DataFrame element compared with another
    element in the DataFrame (default is the element in the same column
    of the previous row).

    Parameters
    ----------
    periods : int, default 1
        Periods to shift for calculating difference, accepts negative
        values.
    axis : {0 or 'index', 1 or 'columns'}, default 0
        Take difference over rows (0) or columns (1).

    Returns
    -------
    DataFrame

    See Also
    --------
    Series.diff : First discrete difference for a Series.
    DataFrame.pct_change : Percent change over given number of periods.
    DataFrame.shift : Shift index by desired number of periods with an
        optional time freq.

    Notes
    -----
    For boolean dtypes, this uses :meth:`operator.xor` rather than
    :meth:`operator.sub`.

    Examples
    --------
    Difference with previous row

    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'a': [1, 2, 3, 4, 5, 6],
    ...                    'b': [1, 1, 2, 3, 5, 8],
    ...                    'c': [1, 4, 9, 16, 25, 36]})
    >>> df.execute()
       a  b   c
    0  1  1   1
    1  2  1   4
    2  3  2   9
    3  4  3  16
    4  5  5  25
    5  6  8  36

    >>> df.diff().execute()
         a    b     c
    0  NaN  NaN   NaN
    1  1.0  0.0   3.0
    2  1.0  1.0   5.0
    3  1.0  1.0   7.0
    4  1.0  2.0   9.0
    5  1.0  3.0  11.0

    Difference with previous column

    >>> df.diff(axis=1).execute()
        a    b     c
    0 NaN  0.0   0.0
    1 NaN -1.0   3.0
    2 NaN -1.0   7.0
    3 NaN -1.0  13.0
    4 NaN  0.0  20.0
    5 NaN  2.0  28.0

    Difference with 3rd previous row

    >>> df.diff(periods=3).execute()
         a    b     c
    0  NaN  NaN   NaN
    1  NaN  NaN   NaN
    2  NaN  NaN   NaN
    3  3.0  2.0  15.0
    4  3.0  4.0  21.0
    5  3.0  6.0  27.0

    Difference with following row

    >>> df.diff(periods=-1).execute()
         a    b     c
    0 -1.0  0.0  -3.0
    1 -1.0 -1.0  -5.0
    2 -1.0 -1.0  -7.0
    3 -1.0 -2.0  -9.0
    4 -1.0 -3.0 -11.0
    5  NaN  NaN   NaN
    """
    axis = validate_axis(axis, df)
    op = DataFrameDiff(periods=periods, axis=axis)
    return op(df)


def series_diff(series, periods=1):
    """
    First discrete difference of element.
    Calculates the difference of a Series element compared with another
    element in the Series (default is element in previous row).

    Parameters
    ----------
    periods : int, default 1
        Periods to shift for calculating difference, accepts negative
        values.

    Returns
    -------
    Series
        First differences of the Series.

    See Also
    --------
    Series.pct_change :
        Percent change over given number of periods.
    Series.shift :
        Shift index by desired number of periods with an optional time freq.
    DataFrame.diff :
        First discrete difference of object.

    Notes
    -----
    For boolean dtypes, this uses :meth:`operator.xor` rather than
    :meth:`operator.sub`.

    Examples
    --------

    Difference with previous row

    >>> import mars.dataframe as md
    >>> s = md.Series([1, 1, 2, 3, 5, 8])
    >>> s.diff().execute()
    0    NaN
    1    0.0
    2    1.0
    3    1.0
    4    2.0
    5    3.0
    dtype: float64

    Difference with 3rd previous row

    >>> s.diff(periods=3).execute()
    0    NaN
    1    NaN
    2    NaN
    3    2.0
    4    4.0
    5    6.0
    dtype: float64

    Difference with following row

    >>> s.diff(periods=-1).execute()
    0    0.0
    1   -1.0
    2   -1.0
    3   -2.0
    4   -3.0
    5    NaN
    dtype: float64
    """
    op = DataFrameDiff(periods=periods)
    return op(series)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time
import numpy as np
from collections import defaultdict
from typing import Dict

from ....utils import parse_readable_size, create_task_with_error_log
from ..errors import NoAvailableBand
from .manager import SubtaskScheduleInfo

logger = logging.getLogger(__name__)

# the default times for speculative subtask execution.
DEFAULT_SUBTASK_SPECULATION_THRESHOLD = 0.75
DEFAULT_SUBTASK_SPECULATION_INTERVAL = 5  # time unit: seconds
DEFAULT_SUBTASK_SPECULATION_MIN_TASK_RUNTIME = 3
DEFAULT_SUBTASK_SPECULATION_MULTIPLIER = 1.5
DEFAULT_SUBTASK_MAX_CONCURRENT_RUN = 3


class SpeculativeScheduler:
    _grouped_unfinished_subtasks: Dict[
        str, Dict[str, SubtaskScheduleInfo]
    ]  # key is subtask logic key
    _grouped_finished_subtasks: Dict[
        str, Dict[str, SubtaskScheduleInfo]
    ]  # key is subtask logic key

    def __init__(
        self, queueing_ref, global_resource_ref, speculation_config: Dict[str, any]
    ):
        self._grouped_unfinished_subtasks = defaultdict(dict)
        self._grouped_finished_subtasks = defaultdict(dict)
        self._queueing_ref = queueing_ref
        self._global_resource_ref = global_resource_ref
        self._speculation_config = speculation_config
        self._subtask_speculation_enabled = speculation_config.get("enabled", False)
        assert self._subtask_speculation_enabled in (True, False)
        self._subtask_speculation_dry = speculation_config.get("dry", False)
        self._subtask_speculation_threshold = parse_readable_size(
            speculation_config.get("threshold", DEFAULT_SUBTASK_SPECULATION_THRESHOLD)
        )[0]
        self._subtask_speculation_interval = speculation_config.get(
            "interval", DEFAULT_SUBTASK_SPECULATION_INTERVAL
        )
        self._subtask_speculation_min_task_runtime = speculation_config.get(
            "min_task_runtime", DEFAULT_SUBTASK_SPECULATION_MIN_TASK_RUNTIME
        )
        self._subtask_speculation_multiplier = speculation_config.get(
            "multiplier", DEFAULT_SUBTASK_SPECULATION_MULTIPLIER
        )
        self._subtask_speculation_max_concurrent_run = speculation_config.get(
            "max_concurrent_run", DEFAULT_SUBTASK_MAX_CONCURRENT_RUN
        )
        if self._subtask_speculation_enabled:
            assert 1 >= self._subtask_speculation_threshold > 0
            assert self._subtask_speculation_interval > 0
            assert self._subtask_speculation_min_task_runtime > 0
            assert self._subtask_speculation_multiplier > 0
            assert self._subtask_speculation_max_concurrent_run > 0
        self._speculation_execution_task = None

    async def start(self):
        if self._subtask_speculation_enabled:
            self._speculation_execution_task = create_task_with_error_log(
                self._speculative_execution_loop()
            )
            logger.info(
                "Speculative execution started with config %s.",
                self._speculation_config,
            )

    async def stop(self):
        if self._subtask_speculation_enabled:
            self._speculation_execution_task.cancel()
            try:
                await self._speculation_execution_task
            except asyncio.CancelledError:
                pass
            logger.info("Speculative execution stopped.")

    def add_subtask(self, subtask_info: SubtaskScheduleInfo):
        # duplicate subtask add will be handled in `_speculative_execution`.
        subtask = subtask_info.subtask
        self._grouped_unfinished_subtasks[subtask.logic_key][
            subtask.subtask_id
        ] = subtask_info

    def finish_subtask(self, subtask_info: SubtaskScheduleInfo):
        subtask = subtask_info.subtask
        grouped_finished_subtasks = self._grouped_finished_subtasks[subtask.logic_key]
        grouped_finished_subtasks[subtask.subtask_id] = subtask_info
        self._grouped_unfinished_subtasks[subtask.logic_key].pop(
            subtask.subtask_id, None
        )
        if len(grouped_finished_subtasks) == subtask.logic_parallelism:
            self._grouped_finished_subtasks.pop(subtask.logic_key)
            self._grouped_unfinished_subtasks.pop(subtask.logic_key, None)
            logger.info(
                "Subtask group with logic key %s parallelism %s finished.",
                subtask.logic_key,
                subtask.logic_parallelism,
            )

    async def _speculative_execution_loop(self):
        while True:
            # check subtasks in the same group which has same logic key periodically, if some subtasks hasn't been
            # finished in a considerably longer duration, then those subtasks maybe slow/hang subtasks, try resubmit
            # it to other bands too.
            await asyncio.sleep(self._subtask_speculation_interval)
            await self._speculative_execution()

    async def _speculative_execution(self):
        for logic_key, subtask_infos_dict in dict(
            self._grouped_finished_subtasks
        ).items():
            if not subtask_infos_dict:  # pragma: no cover
                continue
            subtask_infos = subtask_infos_dict.values()
            one_subtask = next(iter(subtask_infos)).subtask
            parallelism = one_subtask.logic_parallelism
            spec_threshold = max(
                1, int(self._subtask_speculation_threshold * parallelism)
            )
            # if finished subtasks reached the spec_threshold, try to find slow/hang unfinished subtasks
            if parallelism > len(subtask_infos) >= spec_threshold:
                unfinished_subtask_infos = self._grouped_unfinished_subtasks[
                    logic_key
                ].values()
                # sort finished subtasks by running time
                duration_array = np.sort(
                    np.array(
                        [info.end_time - info.start_time for info in subtask_infos]
                    )
                )
                median = np.percentile(duration_array, 50)
                duration_threshold = max(
                    median * self._subtask_speculation_multiplier,
                    self._subtask_speculation_min_task_runtime,
                )
                now = time.time()
                # find subtasks whose duration is large enough so that can be took as slow/hang subtasks
                unfinished_subtask_infos = [
                    info
                    for info in unfinished_subtask_infos
                    if info not in subtask_infos
                    and now - info.start_time > duration_threshold
                ]
                if not unfinished_subtask_infos:  # pragma: no cover
                    continue
                exclude_bands = set()
                for info in unfinished_subtask_infos:
                    exclude_bands.update(info.band_futures.keys())
                remaining_resources = (
                    await self._global_resource_ref.get_remaining_resources()
                )
                logger.warning(
                    "%s subtasks in %s for group %s has not been finished in %s seconds on bands %s, "
                    "median duration is %s, average duration for %s finished subtasks "
                    "is %s. trying speculative running. "
                    "Current cluster remaining resources %s",
                    len(unfinished_subtask_infos),
                    parallelism,
                    logic_key,
                    duration_threshold,
                    exclude_bands,
                    median,
                    len(subtask_infos),
                    duration_array.mean(),
                    remaining_resources,
                )
                # TODO(chaokunyang) If too many subtasks got stale on same node, mark the node as slow node.
                for subtask_info in unfinished_subtask_infos:
                    subtask = subtask_info.subtask
                    if subtask.retryable:
                        logger.warning(
                            "Subtask %s has not been finished in %s seconds on bands %s, "
                            "trying speculative running.",
                            subtask.subtask_id,
                            now - subtask_info.start_time,
                            list(subtask_info.band_futures.keys()),
                        )
                        await self._submit_speculative_subtask(
                            subtask_info, exclude_bands
                        )
                    else:
                        logger.warning(
                            "Unretryable subtask %s has not been finished in %s seconds "
                            "on bands %s, median duration is %s, it may hang.",
                            subtask.subtask_id,
                            (now - subtask_info.start_time),
                            list(subtask_info.band_futures.keys()),
                            median,
                        )
                await self._queueing_ref.submit_subtasks.tell()

    async def _submit_speculative_subtask(self, subtask_info, exclude_bands):
        subtask = subtask_info.subtask
        if (
            subtask_info.num_speculative_concurrent_run
            == self._subtask_speculation_max_concurrent_run
        ):
            logger.debug(
                "Subtask %s speculative run has reached max limit %s, "
                "won't submit another speculative run.",
                subtask.subtask_id,
                self._subtask_speculation_max_concurrent_run,
            )
            return
        if not self._subtask_speculation_dry:
            if (
                len(subtask_info.band_futures)
                < subtask_info.num_speculative_concurrent_run + 1
            ):
                # ensure same subtask won't be submitted to same worker.
                logger.info(
                    "Speculative execution for subtask %s has not been submitted to worker,"
                    "waiting for being submitted to worker."
                    "Cluster resources may be not enough after excluded %s",
                    subtask.subtask_id,
                    exclude_bands,
                )
                return
            try:
                await self._queueing_ref.add_subtasks(
                    [subtask],
                    [subtask.priority or tuple()],
                    exclude_bands=exclude_bands,
                    random_when_unavailable=False,
                )
                logger.info(
                    "Added subtask %s to queue excluded from %s.",
                    subtask.subtask_id,
                    exclude_bands,
                )
                subtask_info.num_speculative_concurrent_run += 1
                if (
                    subtask_info.num_speculative_concurrent_run
                    == self._subtask_speculation_max_concurrent_run
                ):
                    logger.info(
                        "Subtask %s reached max speculative execution: %s",
                        subtask.subtask_id,
                        self._subtask_speculation_max_concurrent_run,
                    )
            except NoAvailableBand:
                logger.warning(
                    "No bands available for subtask %s after excluded bands %s, "
                    "try resubmit later.",
                    subtask.subtask_id,
                    exclude_bands,
                )
            except KeyError as e:  # pragma: no cover
                # if the subtask happen to be finished, it's input chunk may got gc, if assigning to band
                # needs to know input meta, we'll get KeyError or something else, just ignore it.
                logger.warning(
                    "Subtask %s may happen to be finished just now, cannot add it to "
                    "subtask queue, got error %s, just ignore it.",
                    subtask.subtask_id,
                    e,
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import Int64Field, Float64Field
from ..array_utils import array_module
from ..utils import gen_random_seeds
from .core import TensorRandomOperandMixin, TensorSimpleRandomData


class TensorRandint(TensorSimpleRandomData, TensorRandomOperandMixin):
    _op_type_ = OperandDef.RAND_RANDINT

    _fields_ = "low", "high", "density", "size"
    low = Int64Field("low")
    high = Int64Field("high")
    density = Float64Field("density")
    _func_name = "randint"

    def __call__(self, chunk_size=None):
        return self.new_tensor(None, None, raw_chunk_size=chunk_size)

    @classmethod
    def execute(cls, ctx, op):
        if op.sparse:
            cls.execute_sparse(ctx, op)
        else:
            super().execute(ctx, op)

    @classmethod
    def execute_sparse(cls, ctx, op):
        from ...lib.sparse import SparseNDArray
        from ...lib.sparse.core import cps, sps

        xp = array_module(op.gpu)
        if op.seed:
            rs = np.random.RandomState(op.seed)
        else:
            rs = None

        chunk = op.outputs[0]
        if chunk.ndim > 2:
            raise NotImplementedError

        low = 1 if op.low == 0 else op.low

        rs = rs or xp.random
        size = int(np.ceil(np.prod(chunk.shape) * op.density))
        xps = cps if op.gpu else sps
        ij = xp.empty((2, size))
        ij[0] = rs.randint(chunk.shape[0], size=size)
        ij[1] = rs.randint(chunk.shape[1], size=size)
        data = rs.randint(low, op.high, size=size).astype(op.dtype)
        m = xps.coo_matrix((data, ij), chunk.shape).tocsr()
        m.data[m.data >= op.high] = op.high - 1

        # scipy.sparse is too slow, we remove the precise version due to the performance
        # m = sps.random(*chunk.shape, density=op.density, format='csr')
        # m.data = (rs or xp.random).randint(low, op.high, size=m.data.size)\
        #     .astype(op.dtype)

        ctx[chunk.key] = SparseNDArray(m)

    @classmethod
    def estimate_size(cls, ctx, op):
        chunk = op.outputs[0]
        if not op.sparse or not getattr(op, "_density", None):
            super().estimate_size(ctx, op)
        else:
            # use density to estimate real memory usage
            nbytes = int(chunk.nbytes * getattr(chunk.op, "_density"))
            ctx[chunk.key] = (nbytes, nbytes)


def randint(
    random_state,
    low,
    high=None,
    size=None,
    dtype="l",
    density=None,
    chunk_size=None,
    gpu=None,
):
    """
    Return random integers from `low` (inclusive) to `high` (exclusive).

    Return random integers from the "discrete uniform" distribution of
    the specified dtype in the "half-open" interval [`low`, `high`). If
    `high` is None (the default), then results are from [0, `low`).

    Parameters
    ----------
    low : int
        Lowest (signed) integer to be drawn from the distribution (unless
        ``high=None``, in which case this parameter is one above the
        *highest* such integer).
    high : int, optional
        If provided, one above the largest (signed) integer to be drawn
        from the distribution (see above for behavior if ``high=None``).
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    dtype : dtype, optional
        Desired dtype of the result. All dtypes are determined by their
        name, i.e., 'int64', 'int', etc, so byteorder is not available
        and a specific precision may have different C types depending
        on the platform. The default value is 'np.int'.
    density: float, optional
        if density specified, a sparse tensor will be created
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : int or Tensor of ints
        `size`-shaped tensor of random integers from the appropriate
        distribution, or a single such random int if `size` not provided.

    See Also
    --------
    random.random_integers : similar to `randint`, only for the closed
        interval [`low`, `high`], and 1 is the lowest value if `high` is
        omitted. In particular, this other one is the one to use to generate
        uniformly distributed discrete non-integers.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.random.randint(2, size=10).execute()
    array([1, 0, 0, 0, 1, 1, 0, 0, 1, 0])
    >>> mt.random.randint(1, size=10).execute()
    array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    Generate a 2 x 4 tensor of ints between 0 and 4, inclusive:

    >>> mt.random.randint(5, size=(2, 4)).execute()
    array([[4, 0, 2, 1],
           [3, 2, 2, 0]])
    """
    sparse = bool(density)
    size = random_state._handle_size(size)
    seed = gen_random_seeds(1, random_state.to_numpy())[0]
    op = TensorRandint(
        seed=seed,
        low=low,
        high=high,
        size=size,
        dtype=dtype,
        gpu=gpu,
        sparse=sparse,
        density=density,
    )
    return op(chunk_size=chunk_size)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number
from typing import Dict

from ...driver import BaseActorDriver


class MarsActorDriver(BaseActorDriver):
    @classmethod
    def setup_cluster(cls, address_to_resources: Dict[str, Dict[str, Number]]):
        # nothing need to be done in driver of Mars backend
        pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import itertools
import logging

from collections import defaultdict
from typing import Dict, List, Optional

from .... import oscar as mo
from ...meta.api import MetaAPI
from ...storage.api import StorageAPI
from ..errors import TileableNotTracked

logger = logging.getLogger(__name__)


class LifecycleTrackerActor(mo.Actor):
    _meta_api: MetaAPI

    def __init__(self, session_id: str):
        self._session_id = session_id
        self._tileable_key_to_chunk_keys = dict()
        self._tileable_ref_counts = defaultdict(lambda: 0)
        self._chunk_ref_counts = defaultdict(lambda: 0)

        self._meta_api: Optional[MetaAPI] = None

    async def __post_create__(self):
        self._meta_api = await MetaAPI.create(self._session_id, self.address)

    async def __pre_destroy__(self):
        chunk_keys = [
            chunk_key
            for chunk_key, ref_count in self._chunk_ref_counts.items()
            if ref_count > 0
        ]
        # remove all chunks
        await self._remove_chunks(chunk_keys)

    @staticmethod
    def gen_uid(session_id):
        return f"{session_id}_lifecycle_tracker"

    def _track(self, tileable_key: str, chunk_keys: List[str]):
        if tileable_key not in self._tileable_key_to_chunk_keys:
            self._tileable_key_to_chunk_keys[tileable_key] = []
        chunk_keys_set = set(self._tileable_key_to_chunk_keys[tileable_key])
        incref_chunk_keys = []
        tileable_ref_count = self._tileable_ref_counts.get(tileable_key, 0)
        for chunk_key in chunk_keys:
            if chunk_key in chunk_keys_set:
                continue
            if tileable_ref_count > 0:
                incref_chunk_keys.extend([chunk_key] * tileable_ref_count)
            self._tileable_key_to_chunk_keys[tileable_key].append(chunk_key)
        if incref_chunk_keys:
            self._incref_chunks(incref_chunk_keys)

    @mo.extensible
    async def track(self, tileable_key: str, chunk_keys: List[str]):
        return await asyncio.to_thread(self._track, tileable_key, chunk_keys)

    @classmethod
    def _check_ref_counts(cls, keys: List[str], ref_counts: List[int]):
        if ref_counts is not None and len(keys) != len(ref_counts):
            raise ValueError(
                f"`ref_counts` should have same size as `keys`, expect {len(keys)}, got {len(ref_counts)}"
            )

    def _incref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        counts = counts if counts is not None else itertools.repeat(1)
        for chunk_key, count in zip(chunk_keys, counts):
            self._chunk_ref_counts[chunk_key] += count

    async def incref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        self._check_ref_counts(chunk_keys, counts)
        return await asyncio.to_thread(self._incref_chunks, chunk_keys, counts=counts)

    def _get_remove_chunk_keys(self, chunk_keys: List[str], counts: List[int] = None):
        to_remove_chunk_keys = []
        counts = counts if counts is not None else itertools.repeat(1)
        for chunk_key, count in zip(chunk_keys, counts):
            ref_count = self._chunk_ref_counts[chunk_key]
            ref_count -= count
            assert ref_count >= 0, f"chunk key {chunk_key} will have negative ref count"
            self._chunk_ref_counts[chunk_key] = ref_count
            if ref_count == 0:
                # remove
                to_remove_chunk_keys.append(chunk_key)
        return to_remove_chunk_keys

    async def decref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        self._check_ref_counts(chunk_keys, counts)
        to_remove_chunk_keys = await asyncio.to_thread(
            self._get_remove_chunk_keys, chunk_keys, counts=counts
        )
        # make _remove_chunks release actor lock so that multiple `decref_chunks` can run concurrently.
        yield self._remove_chunks(to_remove_chunk_keys)

    async def _remove_chunks(self, to_remove_chunk_keys: List[str]):
        if not to_remove_chunk_keys:
            return
        # get meta
        logger.debug("Remove chunks %s with a refcount of zero", to_remove_chunk_keys)
        get_metas = []
        for to_remove_chunk_key in to_remove_chunk_keys:
            get_metas.append(
                self._meta_api.get_chunk_meta.delay(
                    to_remove_chunk_key, fields=["bands"], error="ignore"
                )
            )
        metas = await self._meta_api.get_chunk_meta.batch(*get_metas)

        # filter chunks that not exist
        new_to_remove_chunk_keys = []
        new_metas = []
        for to_remove_chunk_key, meta in zip(to_remove_chunk_keys, metas):
            if meta is not None:
                new_to_remove_chunk_keys.append(to_remove_chunk_key)
                new_metas.append(meta)
        to_remove_chunk_keys = new_to_remove_chunk_keys
        metas = new_metas

        all_bands = [meta["bands"] for meta in metas]
        key_to_addresses = dict()
        for to_remove_chunk_key, bands in zip(to_remove_chunk_keys, all_bands):
            key_to_addresses[to_remove_chunk_key] = bands

        # remove data via storage API
        storage_api_to_deletes = defaultdict(list)
        for key, bands in key_to_addresses.items():
            for band in bands:
                # storage API is cached for same arguments
                storage_api = await StorageAPI.create(
                    self._session_id, band[0], band[1]
                )
                storage_api_to_deletes[storage_api].append(
                    storage_api.delete.delay(key, error="ignore")
                )
        await asyncio.gather(
            *[
                storage_api.delete.batch(*deletes)
                for storage_api, deletes in storage_api_to_deletes.items()
            ]
        )

        # delete meta
        delete_metas = []
        for to_remove_chunk_key in to_remove_chunk_keys:
            delete_metas.append(
                self._meta_api.del_chunk_meta.delay(to_remove_chunk_key)
            )
        await self._meta_api.del_chunk_meta.batch(*delete_metas)

    def get_chunk_ref_counts(self, chunk_keys: List[str]) -> List[int]:
        return [self._chunk_ref_counts[chunk_key] for chunk_key in chunk_keys]

    def get_all_chunk_ref_counts(self) -> Dict[str, int]:
        result = dict()
        for chunk_key, ref_count in self._chunk_ref_counts.items():
            if ref_count > 0:
                result[chunk_key] = ref_count
        return result

    def _incref_tileables(self, tileable_keys: List[str], counts: List[int] = None):
        counts = counts if counts is not None else itertools.repeat(1)
        for tileable_key, count in zip(tileable_keys, counts):
            if tileable_key not in self._tileable_key_to_chunk_keys:
                raise TileableNotTracked(f"tileable {tileable_key} not tracked before")
            self._tileable_ref_counts[tileable_key] += count
            incref_chunk_keys = self._tileable_key_to_chunk_keys[tileable_key]
            # incref chunks for this tileable
            logger.debug(
                "Incref chunks %s while increfing tileable %s",
                incref_chunk_keys,
                tileable_key,
            )
            chunk_counts = None if count == 1 else [count] * len(incref_chunk_keys)
            self._incref_chunks(incref_chunk_keys, counts=chunk_counts)

    async def incref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        self._check_ref_counts(tileable_keys, counts)
        return await asyncio.to_thread(
            self._incref_tileables, tileable_keys, counts=counts
        )

    def _get_decref_chunk_keys(
        self, tileable_keys: List[str], counts: List[int] = None
    ) -> Dict[str, int]:
        decref_chunk_keys = dict()
        counts = counts if counts is not None else itertools.repeat(1)
        for tileable_key, count in zip(tileable_keys, counts):
            if tileable_key not in self._tileable_key_to_chunk_keys:
                raise TileableNotTracked(f"tileable {tileable_key} not tracked before")
            self._tileable_ref_counts[tileable_key] -= count

            for chunk_key in self._tileable_key_to_chunk_keys[tileable_key]:
                if chunk_key not in decref_chunk_keys:
                    decref_chunk_keys[chunk_key] = count
                else:
                    decref_chunk_keys[chunk_key] += count
        logger.debug(
            "Decref chunks %s while decrefing tileables %s",
            decref_chunk_keys,
            tileable_keys,
        )
        return decref_chunk_keys

    async def decref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        self._check_ref_counts(tileable_keys, counts)
        decref_chunk_key_to_counts = await asyncio.to_thread(
            self._get_decref_chunk_keys, tileable_keys, counts=counts
        )
        to_remove_chunk_keys = await asyncio.to_thread(
            self._get_remove_chunk_keys,
            list(decref_chunk_key_to_counts),
            counts=list(decref_chunk_key_to_counts.values()),
        )
        # make _remove_chunks release actor lock
        yield self._remove_chunks(to_remove_chunk_keys)

    def get_tileable_ref_counts(self, tileable_keys: List[str]) -> List[int]:
        return [
            self._tileable_ref_counts[tileable_key] for tileable_key in tileable_keys
        ]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import warnings
from math import gcd
from collections import namedtuple
from typing import Callable, Tuple, Union

import numpy as np
from scipy import special
from scipy.stats import distributions

from ... import tensor as mt
from ...core import ExecutableTuple
from ...typing import TileableType


KstestResult = namedtuple("KstestResult", ("statistic", "pvalue"))
Ks_2sampResult = KstestResult


def _compute_prob_inside_method(m, n, g, h):  # pragma: no cover
    """
    Count the proportion of paths that stay strictly inside two diagonal lines.

    Parameters
    ----------
    m : integer
        m > 0
    n : integer
        n > 0
    g : integer
        g is greatest common divisor of m and n
    h : integer
        0 <= h <= lcm(m,n)

    Returns
    -------
    p : float
        The proportion of paths that stay inside the two lines.


    Count the integer lattice paths from (0, 0) to (m, n) which satisfy
    |x/m - y/n| < h / lcm(m, n).
    The paths make steps of size +1 in either positive x or positive y directions.

    We generally follow Hodges' treatment of Drion/Gnedenko/Korolyuk.
    Hodges, J.L. Jr.,
    "The Significance Probability of the Smirnov Two-Sample Test,"
    Arkiv fiur Matematik, 3, No. 43 (1958), 469-86.

    """
    # Probability is symmetrical in m, n.  Computation below uses m >= n.
    if m < n:
        m, n = n, m
    mg = m // g
    ng = n // g

    # Count the integer lattice paths from (0, 0) to (m, n) which satisfy
    # |nx/g - my/g| < h.
    # Compute matrix A such that:
    #  A(x, 0) = A(0, y) = 1
    #  A(x, y) = A(x, y-1) + A(x-1, y), for x,y>=1, except that
    #  A(x, y) = 0 if |x/m - y/n|>= h
    # Probability is A(m, n)/binom(m+n, n)
    # Optimizations exist for m==n, m==n*p.
    # Only need to preserve a single column of A, and only a sliding window of it.
    # minj keeps track of the slide.
    minj, maxj = 0, min(int(np.ceil(h / mg)), n + 1)
    curlen = maxj - minj
    # Make a vector long enough to hold maximum window needed.
    lenA = min(2 * maxj + 2, n + 1)
    # This is an integer calculation, but the entries are essentially
    # binomial coefficients, hence grow quickly.
    # Scaling after each column is computed avoids dividing by a
    # large binomial coefficient at the end, but is not sufficient to avoid
    # the large dynamic range which appears during the calculation.
    # Instead we rescale based on the magnitude of the right most term in
    # the column and keep track of an exponent separately and apply
    # it at the end of the calculation.  Similarly when multiplying by
    # the binomial coefficient
    dtype = np.float64
    A = np.zeros(lenA, dtype=dtype)
    # Initialize the first column
    A[minj:maxj] = 1
    expnt = 0
    for i in range(1, m + 1):
        # Generate the next column.
        # First calculate the sliding window
        lastminj, lastlen = minj, curlen
        minj = max(int(np.floor((ng * i - h) / mg)) + 1, 0)
        minj = min(minj, n)
        maxj = min(int(np.ceil((ng * i + h) / mg)), n + 1)
        if maxj <= minj:
            return 0
        # Now fill in the values
        A[0 : maxj - minj] = np.cumsum(A[minj - lastminj : maxj - lastminj])
        curlen = maxj - minj
        if lastlen > curlen:
            # Set some carried-over elements to 0
            A[maxj - minj : maxj - minj + (lastlen - curlen)] = 0
        # Rescale if the right most value is over 2**900
        val = A[maxj - minj - 1]
        _, valexpt = math.frexp(val)
        if valexpt > 900:
            # Scaling to bring down to about 2**800 appears
            # sufficient for sizes under 10000.
            valexpt -= 800
            A = np.ldexp(A, -valexpt)
            expnt += valexpt

    val = A[maxj - minj - 1]
    # Now divide by the binomial (m+n)!/m!/n!
    for i in range(1, n + 1):
        val = (val * i) / (m + i)
        _, valexpt = math.frexp(val)
        if valexpt < -128:
            val = np.ldexp(val, -valexpt)
            expnt += valexpt
    # Finally scale if needed.
    return np.ldexp(val, expnt)


def _compute_prob_outside_square(n, h):  # pragma: no cover
    """
    Compute the proportion of paths that pass outside the two diagonal lines.

    Parameters
    ----------
    n : integer
        n > 0
    h : integer
        0 <= h <= n

    Returns
    -------
    p : float
        The proportion of paths that pass outside the lines x-y = +/-h.

    """
    # Compute Pr(D_{n,n} >= h/n)
    # Prob = 2 * ( binom(2n, n-h) - binom(2n, n-2a) + binom(2n, n-3a) - ... )  / binom(2n, n)
    # This formulation exhibits subtractive cancellation.
    # Instead divide each term by binom(2n, n), then factor common terms
    # and use a Horner-like algorithm
    # P = 2 * A0 * (1 - A1*(1 - A2*(1 - A3*(1 - A4*(...)))))

    P = 0.0
    k = int(np.floor(n / h))
    while k >= 0:
        p1 = 1.0
        # Each of the Ai terms has numerator and denominator with h simple terms.
        for j in range(h):
            p1 = (n - k * h - j) * p1 / (n + k * h + j + 1)
        P = p1 * (1.0 - P)
        k -= 1
    return 2 * P


def _count_paths_outside_method(m, n, g, h):  # pragma: no cover
    """
    Count the number of paths that pass outside the specified diagonal.

    Parameters
    ----------
    m : integer
        m > 0
    n : integer
        n > 0
    g : integer
        g is greatest common divisor of m and n
    h : integer
        0 <= h <= lcm(m,n)

    Returns
    -------
    p : float
        The number of paths that go low.
        The calculation may overflow - check for a finite answer.

    Raises
    ------
    FloatingPointError: Raised if the intermediate computation goes outside
    the range of a float.

    Notes
    -----
    Count the integer lattice paths from (0, 0) to (m, n), which at some
    point (x, y) along the path, satisfy:
      m*y <= n*x - h*g
    The paths make steps of size +1 in either positive x or positive y directions.

    We generally follow Hodges' treatment of Drion/Gnedenko/Korolyuk.
    Hodges, J.L. Jr.,
    "The Significance Probability of the Smirnov Two-Sample Test,"
    Arkiv fiur Matematik, 3, No. 43 (1958), 469-86.

    """
    # Compute #paths which stay lower than x/m-y/n = h/lcm(m,n)
    # B(x, y) = #{paths from (0,0) to (x,y) without previously crossing the boundary}
    #         = binom(x, y) - #{paths which already reached the boundary}
    # Multiply by the number of path extensions going from (x, y) to (m, n)
    # Sum.

    # Probability is symmetrical in m, n.  Computation below assumes m >= n.
    if m < n:
        m, n = n, m
    mg = m // g
    ng = n // g

    # Not every x needs to be considered.
    # xj holds the list of x values to be checked.
    # Wherever n*x/m + ng*h crosses an integer
    lxj = n + (mg - h) // mg
    xj = [(h + mg * j + ng - 1) // ng for j in range(lxj)]
    # B is an array just holding a few values of B(x,y), the ones needed.
    # B[j] == B(x_j, j)
    if lxj == 0:
        return np.round(special.binom(m + n, n))
    B = np.zeros(lxj)
    B[0] = 1
    # Compute the B(x, y) terms
    # The binomial coefficient is an integer, but special.binom() may return a float.
    # Round it to the nearest integer.
    for j in range(1, lxj):
        Bj = np.round(special.binom(xj[j] + j, j))
        if not np.isfinite(Bj):
            raise FloatingPointError()
        for i in range(j):
            bin = np.round(
                special.binom(xj[j] - xj[i] + j - i, j - i)
            )  # pylint: disable=redefined-builtin
            Bj -= bin * B[i]
        B[j] = Bj
        if not np.isfinite(Bj):
            raise FloatingPointError()
    # Compute the number of path extensions...
    num_paths = 0
    for j in range(lxj):
        bin = np.round(special.binom((m - xj[j]) + (n - j), n - j))
        term = B[j] * bin
        if not np.isfinite(term):
            raise FloatingPointError()
        num_paths += term
    return np.round(num_paths)


def _attempt_exact_2kssamp(n1, n2, g, d, alternative):  # pragma: no cover
    """Attempts to compute the exact 2sample probability.

    n1, n2 are the sample sizes
    g is the gcd(n1, n2)
    d is the computed max difference in ECDFs

    Returns (success, d, probability)
    """
    lcm = (n1 // g) * n2
    h = int(np.round(d * lcm))
    d = h * 1.0 / lcm
    if h == 0:
        return True, d, 1.0
    saw_fp_error, prob = False, np.nan
    try:
        if alternative == "two-sided":
            if n1 == n2:
                prob = _compute_prob_outside_square(n1, h)
            else:
                prob = 1 - _compute_prob_inside_method(n1, n2, g, h)
        else:
            if n1 == n2:
                # prob = binom(2n, n-h) / binom(2n, n)
                # Evaluating in that form incurs roundoff errors
                # from special.binom. Instead calculate directly
                jrange = np.arange(h)
                prob = np.prod((n1 - jrange) / (n1 + jrange + 1.0))
            else:
                num_paths = _count_paths_outside_method(n1, n2, g, h)
                bin = special.binom(n1 + n2, n1)  # pylint: disable=redefined-builtin
                if (
                    not np.isfinite(bin)
                    or not np.isfinite(num_paths)
                    or num_paths > bin
                ):
                    saw_fp_error = True
                else:
                    prob = num_paths / bin

    except FloatingPointError:
        saw_fp_error = True

    if saw_fp_error:
        return False, d, np.nan
    if not (0 <= prob <= 1):
        return False, d, prob
    return True, d, prob


def _calc_prob_2samp(d, n1, n2, alternative, mode):  # pragma: no cover
    MAX_AUTO_N = 10000  # 'auto' will attempt to be exact if n1,n2 <= MAX_AUTO_N

    g = gcd(n1, n2)
    n1g = n1 // g
    n2g = n2 // g
    prob = -mt.inf
    original_mode = mode
    if mode == "auto":
        mode = "exact" if max(n1, n2) <= MAX_AUTO_N else "asymp"
    elif mode == "exact":
        # If lcm(n1, n2) is too big, switch from exact to asymp
        if n1g >= np.iinfo(np.int_).max / n2g:
            mode = "asymp"
            warnings.warn(
                f"Exact ks_2samp calculation not possible with samples sizes "
                f"{n1} and {n2}. Switching to 'asymp'.",
                RuntimeWarning,
            )

    if mode == "exact":
        success, d, prob = _attempt_exact_2kssamp(n1, n2, g, d, alternative)
        if not success:
            mode = "asymp"
            if original_mode == "exact":
                warnings.warn(
                    f"ks_2samp: Exact calculation unsuccessful. "
                    f"Switching to mode={mode}.",
                    RuntimeWarning,
                )

    if mode == "asymp":
        # The product n1*n2 is large.  Use Smirnov's asymptotic formula.
        # Ensure float to avoid overflow in multiplication
        # sorted because the one-sided formula is not symmetric in n1, n2
        m, n = sorted([float(n1), float(n2)], reverse=True)
        en = m * n / (m + n)
        if alternative == "two-sided":
            prob = distributions.kstwo.sf(d, np.round(en))
        else:
            z = np.sqrt(en) * d
            # Use Hodges' suggested approximation Eqn 5.3
            # Requires m to be the larger of (n1, n2)
            expt = -2 * z**2 - 2 * z * (m + 2 * n) / np.sqrt(m * n * (m + n)) / 3.0
            prob = np.exp(expt)

    return np.clip(prob, 0, 1)


def _compute_dplus(cdfvals, n):
    """Computes D+ as used in the Kolmogorov-Smirnov test.

    Parameters
    ----------
    cdfvals: array_like
      Sorted array of CDF values between 0 and 1

    Returns
    -------
      Maximum distance of the CDF values below Uniform(0, 1)
    """
    return (mt.arange(1.0, n + 1) / n - cdfvals).max()


def _compute_dminus(cdfvals, n):
    """Computes D- as used in the Kolmogorov-Smirnov test.

    Parameters
    ----------
    cdfvals: array_like
      Sorted array of CDF values between 0 and 1

    Returns
    -------
      Maximum distance of the CDF values above Uniform(0, 1)
    """
    return (cdfvals - mt.arange(0.0, n) / n).max()


def ks_1samp(
    x: Union[np.ndarray, list, TileableType],
    cdf: Callable,
    args: Tuple = (),
    alternative: str = "two-sided",
    mode: str = "auto",
):
    """
    Performs the one-sample Kolmogorov-Smirnov test for goodness of fit.

    This test compares the underlying distribution F(x) of a sample
    against a given continuous distribution G(x). See Notes for a description
    of the available null and alternative hypotheses.

    Parameters
    ----------
    x : array_like
        a 1-D array of observations of iid random variables.
    cdf : callable
        callable used to calculate the cdf.
    args : tuple, sequence, optional
        Distribution parameters, used with `cdf`.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Defines the null and alternative hypotheses. Default is 'two-sided'.
        Please see explanations in the Notes below.
    mode : {'auto', 'exact', 'approx', 'asymp'}, optional
        Defines the distribution used for calculating the p-value.
        The following options are available (default is 'auto'):

          * 'auto' : selects one of the other options.
          * 'exact' : uses the exact distribution of test statistic.
          * 'approx' : approximates the two-sided probability with twice
            the one-sided probability
          * 'asymp': uses asymptotic distribution of test statistic

    Returns
    -------
    statistic : float
        KS test statistic, either D, D+ or D- (depending on the value
        of 'alternative')
    pvalue :  float
        One-tailed or two-tailed p-value.

    See Also
    --------
    ks_2samp, kstest

    Notes
    -----
    There are three options for the null and corresponding alternative
    hypothesis that can be selected using the `alternative` parameter.

    - `two-sided`: The null hypothesis is that the two distributions are
      identical, F(x)=G(x) for all x; the alternative is that they are not
      identical.

    - `less`: The null hypothesis is that F(x) >= G(x) for all x; the
      alternative is that F(x) < G(x) for at least one x.

    - `greater`: The null hypothesis is that F(x) <= G(x) for all x; the
      alternative is that F(x) > G(x) for at least one x.

    Note that the alternative hypotheses describe the *CDFs* of the
    underlying distributions, not the observed values. For example,
    suppose x1 ~ F and x2 ~ G. If F(x) > G(x) for all x, the values in
    x1 tend to be less than those in x2.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy import stats
    >>> import mars.tensor as mt
    >>> from mars.tensor.stats import ks_1samp

    >>> np.random.seed(12345678)  #fix random seed to get the same result
    >>> x = mt.linspace(-15, 15, 9, chunk_size=5)
    >>> ks_1samp(x, stats.norm.cdf).execute()
    (0.44435602715924361, 0.038850142705171065)

    >>> ks_1samp(stats.norm.rvs(size=100), stats.norm.cdf).execute()
    KstestResult(statistic=0.165471391799..., pvalue=0.007331283245...)

    *Test against one-sided alternative hypothesis*

    Shift distribution to larger values, so that `` CDF(x) < norm.cdf(x)``:

    >>> x = stats.norm.rvs(loc=0.2, size=100)
    >>> ks_1samp(x, stats.norm.cdf, alternative='less').execute()
    KstestResult(statistic=0.235488541678..., pvalue=1.158315030683...)

    Reject null hypothesis in favor of alternative hypothesis: less

    >>> ks_1samp(x, stats.norm.cdf, alternative='greater').execute()
    KstestResult(statistic=0.010167165616..., pvalue=0.972494973653...)

    Reject null hypothesis in favor of alternative hypothesis: greater

    >>> ks_1samp(x, stats.norm.cdf).execute()
    KstestResult(statistic=0.235488541678..., pvalue=2.316630061366...)

    Don't reject null hypothesis in favor of alternative hypothesis: two-sided

    *Testing t distributed random variables against normal distribution*

    With 100 degrees of freedom the t distribution looks close to the normal
    distribution, and the K-S test does not reject the hypothesis that the
    sample came from the normal distribution:

    >>> ks_1samp(stats.t.rvs(100, size=100), stats.norm.cdf).execute()
    KstestResult(statistic=0.077844250253..., pvalue=0.553155412513...)

    With 3 degrees of freedom the t distribution looks sufficiently different
    from the normal distribution, that we can reject the hypothesis that the
    sample came from the normal distribution at the 10% level:

    >>> ks_1samp(stats.t.rvs(3, size=100), stats.norm.cdf).execute()
    KstestResult(statistic=0.118967105356..., pvalue=0.108627114578...)
    """
    alternative = {"t": "two-sided", "g": "greater", "l": "less"}.get(
        alternative.lower()[0], alternative
    )
    if alternative not in ["two-sided", "greater", "less"]:
        raise ValueError("Unexpected alternative %s" % alternative)

    x = mt.asarray(x)
    N = x.shape[0]
    x = mt.sort(x)
    cdfvals = x.map_chunk(cdf, args=args, elementwise=True)

    if alternative == "greater":
        Dplus = _compute_dplus(cdfvals, N)
        return ExecutableTuple(
            KstestResult(Dplus, Dplus.map_chunk(distributions.ksone.sf, args=(N,)))
        )

    if alternative == "less":
        Dminus = _compute_dminus(cdfvals, N)
        return ExecutableTuple(
            KstestResult(Dminus, Dminus.map_chunk(distributions.ksone.sf, args=(N,)))
        )

    # alternative == 'two-sided':
    Dplus = _compute_dplus(cdfvals, N)
    Dminus = _compute_dminus(cdfvals, N)
    D = mt.stack([Dplus, Dminus]).max()
    if mode == "auto":  # Always select exact
        mode = "exact"
    if mode == "exact":
        prob = D.map_chunk(distributions.kstwo.sf, args=(N,), elementwise=True)
    elif mode == "asymp":
        prob = (D * np.sqrt(N)).map_chunk(distributions.kstwobign.sf, elementwise=True)
    else:
        # mode == 'approx'
        prob = 2 * D.map_chunk(distributions.ksone.sf, args=(N,), elementwise=True)
    prob = mt.clip(prob, 0, 1)
    return ExecutableTuple(KstestResult(D, prob))


def ks_2samp(
    data1: Union[np.ndarray, list, TileableType],
    data2: Union[np.ndarray, list, TileableType],
    alternative: str = "two-sided",
    mode: str = "auto",
):
    """
    Compute the Kolmogorov-Smirnov statistic on 2 samples.

    This is a two-sided test for the null hypothesis that 2 independent samples
    are drawn from the same continuous distribution.  The alternative hypothesis
    can be either 'two-sided' (default), 'less' or 'greater'.

    Parameters
    ----------
    data1, data2 : array_like, 1-Dimensional
        Two arrays of sample observations assumed to be drawn from a continuous
        distribution, sample sizes can be different.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Defines the alternative hypothesis.
        The following options are available (default is 'two-sided'):

          * 'two-sided'
          * 'less': one-sided, see explanation in Notes
          * 'greater': one-sided, see explanation in Notes
    mode : {'auto', 'exact', 'asymp'}, optional
        Defines the method used for calculating the p-value.
        The following options are available (default is 'auto'):

          * 'auto' : use 'exact' for small size arrays, 'asymp' for large
          * 'exact' : use exact distribution of test statistic
          * 'asymp' : use asymptotic distribution of test statistic

    Returns
    -------
    statistic : float
        KS statistic.
    pvalue : float
        Two-tailed p-value.

    See Also
    --------
    kstest, ks_1samp, epps_singleton_2samp, anderson_ksamp

    Notes
    -----
    This tests whether 2 samples are drawn from the same distribution. Note
    that, like in the case of the one-sample KS test, the distribution is
    assumed to be continuous.

    In the one-sided test, the alternative is that the empirical
    cumulative distribution function F(x) of the data1 variable is "less"
    or "greater" than the empirical cumulative distribution function G(x)
    of the data2 variable, ``F(x)<=G(x)``, resp. ``F(x)>=G(x)``.

    If the KS statistic is small or the p-value is high, then we cannot
    reject the hypothesis that the distributions of the two samples
    are the same.

    If the mode is 'auto', the computation is exact if the sample sizes are
    less than 10000.  For larger sizes, the computation uses the
    Kolmogorov-Smirnov distributions to compute an approximate value.

    The 'two-sided' 'exact' computation computes the complementary probability
    and then subtracts from 1.  As such, the minimum probability it can return
    is about 1e-16.  While the algorithm itself is exact, numerical
    errors may accumulate for large sample sizes.   It is most suited to
    situations in which one of the sample sizes is only a few thousand.

    We generally follow Hodges' treatment of Drion/Gnedenko/Korolyuk [1]_.

    References
    ----------
    .. [1] Hodges, J.L. Jr.,  "The Significance Probability of the Smirnov
           Two-Sample Test," Arkiv fiur Matematik, 3, No. 43 (1958), 469-86.


    Examples
    --------
    >>> import numpy as np
    >>> from scipy import stats
    >>> import mars.tensor as mt
    >>> from mars.tensor.stats import ks_2samp
    >>> np.random.seed(12345678)  #fix random seed to get the same result
    >>> n1 = 200  # size of first sample
    >>> n2 = 300  # size of second sample

    For a different distribution, we can reject the null hypothesis since the
    pvalue is below 1%:

    >>> rvs1 = stats.norm.rvs(size=n1, loc=0., scale=1)
    >>> rvs2 = stats.norm.rvs(size=n2, loc=0.5, scale=1.5)
    >>> ks_2samp(rvs1, rvs2).execute()
    KstestResult(statistic=0.20833333333333337, pvalue=5.1292795978041816e-05)

    For a slightly different distribution, we cannot reject the null hypothesis
    at a 10% or lower alpha since the p-value at 0.144 is higher than 10%

    >>> rvs3 = stats.norm.rvs(size=n2, loc=0.01, scale=1.0)
    >>> ks_2samp(rvs1, rvs3).execute()
    KstestResult(statistic=0.10333333333333333, pvalue=0.14691437867433788)

    For an identical distribution, we cannot reject the null hypothesis since
    the p-value is high, 41%:

    >>> rvs4 = stats.norm.rvs(size=n2, loc=0.0, scale=1.0)
    >>> ks_2samp(rvs1, rvs4).execute()
    KstestResult(statistic=0.07999999999999996, pvalue=0.4115432028915931)

    """

    if mode not in ["auto", "exact", "asymp"]:
        raise ValueError(f"Invalid value for mode: {mode}")
    alternative = {"t": "two-sided", "g": "greater", "l": "less"}.get(
        alternative.lower()[0], alternative
    )
    if alternative not in ["two-sided", "less", "greater"]:
        raise ValueError(f"Invalid value for alternative: {alternative}")
    data1 = mt.asarray(data1)
    data2 = mt.asarray(data2)
    data1 = mt.sort(data1)
    data2 = mt.sort(data2)
    n1 = data1.shape[0]
    n2 = data2.shape[0]
    if min(n1, n2) == 0:
        raise ValueError("Data passed to ks_2samp must not be empty")

    data_all = mt.concatenate([data1, data2])
    # using searchsorted solves equal data problem
    cdf1 = mt.searchsorted(data1, data_all, side="right") / n1
    cdf2 = mt.searchsorted(data2, data_all, side="right") / n2
    cddiffs = cdf1 - cdf2
    minS = mt.clip(-mt.min(cddiffs), 0, 1)  # Ensure sign of minS is not negative.
    maxS = mt.max(cddiffs)
    alt2Dvalue = {"less": minS, "greater": maxS, "two-sided": mt.maximum(minS, maxS)}
    d = alt2Dvalue[alternative]
    prob = d.map_chunk(
        _calc_prob_2samp,
        args=(n1, n2, alternative, mode),
        elementwise=True,
        dtype=d.dtype,
    )

    return ExecutableTuple(Ks_2sampResult(d, prob))
